"""Functions for general use"""
######### Package Imports #########################################################################

import os, zipfile, subprocess, uuid, shutil, threading, queue
import pandas as pd
from subprocess import run, PIPE
from functools import partial
from threading import Thread
from pySIMsalabim.utils.general import *
from pySIMsalabim.utils.device_parameters import *
if os.name == 'nt':
    from pySIMsalabim.aux_funcs.PathChecksWin import convert_to_long_path

######### Function Definitions ####################################################################
def parallel_error_message(errorcode):
    """When a 'standard Pascal' fatal error occurs, add the standard error message to be used with parallel which does not read the 

    Parameters
    ----------
    errorcode : int
        the error code

    Returns
    -------
    str
        the error message

    """
    message = ''
    if errorcode >= 90 and errorcode < 100:
        message = 'Error '+str(errorcode) +': '
        if errorcode == 90:
            message += 'Device parameter file corrupted.'
        elif errorcode == 91:
            message += 'Invalid input (physics, or voltage in tVGFile too large).'
        elif errorcode == 92:
            message += 'Invalid input from command line.'
        elif errorcode == 93:
            message += 'Numerical failure.'
        elif errorcode == 94:
            message += 'Failed to converge, halt (FailureMode = 0).'
        elif errorcode == 95:
            message += ' Failed to converge at least 1 point, not halt (FailureMode != 0).'
        elif errorcode == 96:
            message += 'Missing input file.'
        elif errorcode == 97:
            message += 'Runtime exceeds limit set by timeout.'
        elif errorcode == 99:
            message += 'Programming error (i.e. not due to the user!).'
    elif errorcode > 100:
        message = 'Fatal error '+str(errorcode) +': '
        if errorcode == 106:
            message += 'Invalid numeric format: Reported when a non-numeric value is read from a text file.'
        elif errorcode == 200:
            message += 'Division by zero: The application attempted to divide a number by zero.'
        elif errorcode == 201:
            message += 'Range check error.'
        elif errorcode == 202:
            message += 'Stack overflow error: This error is only reported when stack checking is enabled.'
        elif errorcode == 205:
            message += 'Floating point overflow.'
        elif errorcode == 206:
            message = 'Floating point underflow.'
        else:
            message = 'Unknown error code '+str(errorcode) + ' occurred.'

    else:
        message = 'Unknown error code '+str(errorcode) + ' occurred.'

    return message

def run_simulation_parallel(sim_type, cmd_pars_list, session_path, max_jobs = max(1,os.cpu_count()-1), verbose=False, **kwargs):
    """Run the SIMsalabim simulation executable with the chosen device parameters.  
    Select the correct function to run the simulation in parallel based on the operating system.

    Parameters
    ----------
    sim_type : string
        Which type of simulation to run: simss or zimt
    cmd_pars_list : List
        List of list with parameters to add to the simss/zimt cmd line. Each parameter is a dict with par,val keys. 
        Note: when relevant the first entry must be the deviceparameters file with a key: dev_par_file
    session_path : string
        File path of the simss or zimt executable 
    max_jobs : int
        Maximum number of parallel jobs to run. Default is the number of CPU cores - 1
    verbose : bool
        If True, print the output of the simulation to the console
    **kwargs : dict
        Additional keyword arguments to pass to the function
    Returns
    -------
    CompletedProcess
        Output object of with returncode and console output of the simulation

    """    
    force_multithreading = kwargs.get('force_multithreading', False)

    if os.name == 'nt':
        # Windows
        result, msg_list, return_code_list = run_simulation_multithreaded_windows(sim_type, cmd_pars_list, session_path, max_jobs, verbose)
        result_list = return_code_list
    else:
        # Linux
        if shutil.which('parallel') is not None and not force_multithreading:
            result, msg_list, return_code_list  = run_simulation_GNU_parallel(sim_type, cmd_pars_list, session_path, max_jobs, verbose)
            result_list = return_code_list           
        else:
            result, msg_list, return_code_list = run_simulation_multithreaded_linux(sim_type, cmd_pars_list, session_path, max_jobs, verbose)
            result_list = return_code_list

    return result_list

def run_simulation_GNU_parallel(sim_type, cmd_pars_list, session_path, max_jobs = max(1,os.cpu_count()-1),verbose=False):
    """Run the SIMsalabim simulation executable with the chosen device parameters.  
        The simulation is run in parallel using the GNU Parallel program. (https://www.gnu.org/software/parallel/).
        If this command is used please cite:
        Tange, O. (2021, August 22). GNU Parallel 20210822 ('Kabul').
        Zenodo. https://doi.org/10.5281/zenodo.5233953

        To Install GNU Parallel on linux: (not available on Windows)
        sudo apt-get update
        sudo apt-get install parallel
        
        or try to run the install_GNU_parallel_Linux() function from the install module.
        import pySIMsalabim.install as install
        install.install_GNU_parallel_Linux()

        Return the complete result object of the process accompanied by a message with information, 
        in case of both success and failure.

    Parameters
    ----------
    sim_type : string
        Which type of simulation to run: simss or zimt
    cmd_pars_list : List
        List of list with parameters to add to the simss/zimt cmd line. Each parameter is a dict with par,val keys. 
        Note: when relevant the first entry must be the deviceparameters file with a key: dev_par_file
    session_path : string
        File path of the simss or zimt executable 
    max_jobs : int
        Maximum number of parallel jobs to run. Default is the number of CPU cores - 1

    Returns
    -------
    CompletedProcess
        Output object of with returncode and console output of the simulation
    List
        Return list of messages for each simulation
    List
        Return list of return codes for each simulation
    """
    # Construct the command to run the executable
    cmd_line_list = []
    for cmd_pars in cmd_pars_list:
        cmd_line = construct_cmd(sim_type, cmd_pars)
        cmd_line_list.append(cmd_line)
    
   
    # Construct the file and command to run the GNU parallel
    uuid_str = str(uuid.uuid4())
    filename = 'Str4Parallel_'+uuid_str+'.txt'
    log_file = os.path.join(session_path,'logjob_'+uuid_str+ '.dat')

    with open(os.path.join(session_path,filename),'w') as tempfilepar:
        for cmd_line in cmd_line_list:
            tempfilepar.write(cmd_line+'\n')

    cmd_parallel = 'parallel --joblog '+ log_file +' --jobs '+str(int(max_jobs))+' -a '+os.path.join(session_path,filename)
    
    result = run([cmd_parallel], cwd=session_path,stdout=PIPE, check=False, shell=True)
    msg_list,return_code_list = [],[]

    # if result.returncode != 0:
    log = pd.read_csv(log_file, sep='\t',usecols=['Exitval'],on_bad_lines='skip')

    # check if all jobs have been completed successfully, i.e. all exitvals are 0, 95 or 3
    if not all(val in [0, 95, 3] for val in log['Exitval']):
        for idx, val in enumerate(log['Exitval']):
            message = ''
            if val != 0 and val != 95 and val != 3:
                if val == 91:
                    # look for the logfile to get more information about the error
                    for c in cmd_pars_list[idx]:
                        if c['par'] == 'logFile':
                            logFile = c['val']
                            break
                    if os.path.isfile(os.path.join(session_path,logFile)):
                        # find line with 'Program will be terminated.' and store the next lines as the error message
                        startMessage = False
                        with open(os.path.join(session_path,logFile),'r') as f:
                            for line_console in f:
                                if startMessage is True:
                                    # The actual error message. Since the error message can be multi-line, append each line.
                                    message = message + line_console + '\n'
                                if 'Program will be terminated.' in line_console:
                                    # Last 'regular' line of the console output. The next line is from the error message.
                                    startMessage = True
                    if message == '':
                        message = 'Simulation raised an error with Errorcode: ' + str(val) + '\n\n' + parallel_error_message(val)
                elif val == 217:
                    print(cmd_line_list[idx])
                elif val >= 90:
                    # Show the message as an error on the screen. Do not continue to the simulation results page.
                    msg_list.append('Simulation raised an error with Errorcode: ' + str(val) + '\n\n' + parallel_error_message(val))
                
                else:
                    msg_list.append(parallel_error_message(val))
            else:
                if val == 95:
                    # In case of errorcode 95, failures during the simulations were encountered but the simulation did not halt. Show 'error' messages on the UI.
                    msg_list.append('Simulation completed but raised errorcode: ' + str(val) + '\n\n' + 'The simulation finished but at least 1 point did not converge.')
                elif val == 3:
                    # Special case, should not occur in the web version.
                    # When the program exits as a success but no simulation has been run, e.g. in the case of the autotidy functionality. 
                    msg_list.append('Action completed')
                else:
                    # Simulation completed as expected.
                    msg_list.append('Simulation completed.')

    return_code_list = log['Exitval'].tolist()  

    # remove the temporary files
    os.remove(os.path.join(session_path,filename))
    os.remove(log_file)

    return result, msg_list, return_code_list

def run_simulation_multithreaded_windows(sim_type,cmd_pars_list,session_path,max_jobs=max(1,os.cpu_count()-1),verbose=False):
    """Runs simulations in parallel on max_jobs number of threads.  
    This procedure should work on Windows and Linux but it is not as efficient as run_parallel_simu on Linux.
    Yet, it is the only way to run simulations in parallel on Windows in a thread safe way and making sure that two thread do not try to write to the same file at the same time.
    This is achieved by running the simulation in a temporary folder with a copy of all the necessary input files and then moving the output files to the original folder.


    Parameters
    ----------
    sim_type : string
        Which type of simulation to run: simss or zimt
    cmd_pars_list : List
        List of list with parameters to add to the simss/zimt cmd line. Each parameter is a dict with par,val keys. 
        Note: when relevant the first entry must be the deviceparameters file with a key: dev_par_file
    session_path : string
        File path of the simss or zimt executable 
    max_jobs : int
        Maximum number of parallel jobs to run. Default is the number of CPU cores - 1
    verbose : bool
        If True, print the output of the simulation to the console
    
    Returns
    -------
    List
        List of CompletedProcess objects with returncode and console output of the simulation

    """    
    
    lock = threading.Lock()    # Create a lock   
    semaphore = threading.Semaphore(max_jobs)  # Create a semaphore

    # Create tmp folder
    rnd_ID = str(uuid.uuid4())
    tmp_folder = os.path.join(session_path, 'tmp'+rnd_ID) # Generate a unique ID for the temporary folder to make sure there are no conflicts with other processes
    if os.path.isdir(tmp_folder):
        shutil.rmtree(tmp_folder)
        os.mkdir(tmp_folder)
    else:
        os.mkdir(tmp_folder)

    tmp_folder_lst = [tmp_folder] * len(cmd_pars_list)

    # Create a queue to communicate with the worker threads
    q = queue.Queue()

    # Start worker threads
    threads = []
    for i in range(len(cmd_pars_list)):
        t = CustomThread(target=partial(worker_windows,q=q,lock=lock,tmp_folder=tmp_folder_lst[i],semaphore=semaphore,verbose=verbose))
        t.start()
        threads.append(t)

    # Add tasks to the queue
    for code_name, cmd_pars, path in zip([sim_type] * len(cmd_pars_list), cmd_pars_list, tmp_folder_lst):
        q.put((code_name, cmd_pars, path, session_path))

    # Wait for all tasks to be finished
    q.join()
    
    # Stop workers
    for i in range(len(cmd_pars_list)):
        q.put(None)
    
    result_list = []
    for t in threads:
        result_list.append(t.join())
     
    # # Clean up
    shutil.rmtree(tmp_folder)

    return_code_list = [res[0].returncode for res in result_list]
    message_list = []
    for res in return_code_list:
        if res == 91:
            # look for the logfile to get more information about the error
            for cmd_pars in cmd_pars_list:
                for c in cmd_pars:
                    if c['par'] == 'logFile':
                        logFile = c['val']
                        break
            if os.path.isfile(os.path.join(session_path,logFile)):
                # find line with 'Program will be terminated.' and store the next lines as the error message
                startMessage = False
                message = ''
                with open(os.path.join(session_path,logFile),'r') as f:
                    for line_console in f:
                        if startMessage is True:
                            # The actual error message. Since the error message can be multi-line, append each line.
                            message = message + line_console + '\n'
                        if 'Program will be terminated.' in line_console:
                            # Last 'regular' line of the console output. The next line is from the error message.
                            startMessage = True
                if message == '':
                    message = 'Simulation raised an error with Errorcode: ' + str(res) + '\n\n' + parallel_error_message(res)
                message_list.append(message)
            else:
                message_list.append('Simulation raised an error with Errorcode: ' + str(res) + '\n\n' + parallel_error_message(res))
        else:
            message_list.append(parallel_error_message(res))

    # message_list = [parallel_error_message(res) for res in return_code_list]
    if not all(val in [0, 95, 3] for val in return_code_list):
        # check if only one error code different from 0, 95 or 3
        if len(set([val for val in return_code_list if val not in [0, 95, 3]])) == 1:
            # if so, return that error code and the corresponding message
            idx = return_code_list.index([val for val in return_code_list if val not in [0, 95, 3]][0])
            result = return_code_list[idx]
            message = message_list[idx]
            result_list = [(result, message)]
        else:
            print(f"Multiple different errors occurred during the parallel simulations: {set([val for val in return_code_list if val not in [0, 95, 3]])}. Returning error code 666.")
            result = 666
    elif all(val in [0, 95] for val in return_code_list):
        result = 95
    elif all(val in [0, 3] for val in return_code_list):
        result = 3
    else:
        result = 0
    return result, message_list, return_code_list

def worker_windows(q,lock,tmp_folder,semaphore,verbose=False):
    """Worker function that runs the simulation in a temporary folder and moves the output files to the original folder. 

    Parameters
    ----------
    q : queue
        Queue to communicate with the main thread
    lock : threading.Lock
        Lock to prevent multiple threads from writing to the same file at the same time
    tmp_folder : string
        Temporary folder to run the simulation in
    semaphore : threading.Semaphore
        Semaphore to limit the number of parallel jobs
    verbose : bool
        If True, print the output of the simulation to the console

    """
    while True:
        # Get task from the queue
        task = q.get()
        if task is None:
            break

        # Unpack task
        sim_type, cmd_pars, path, session_path = task

        device_parameters = None

        for cmd_par in cmd_pars:
            if cmd_par['par'] == 'dev_par_file':
                device_parameters = cmd_par['val']
                break
    
        if device_parameters is None:
            raise ValueError('Device parameters file not found in the command parameters list.')

        # Acquire semaphore
        semaphore.acquire()

        # create a temporary folder for the simulation
        ID = str(uuid.uuid4()) # Generate a unique ID for the temporary folder
        tmp_folder = os.path.join(tmp_folder, ID)
        if not os.path.isdir(tmp_folder):
            os.mkdir(tmp_folder)

        # Copy all necessary files to the temporary folder
        lock.acquire()
        # copy the executable to the temporary folder

        if os.name == 'nt':
            sim_exe_path = convert_to_long_path(os.path.join(session_path, sim_type + '.exe'))
            shutil.copy(sim_exe_path, tmp_folder)
        else:
            shutil.copy(os.path.join(session_path, sim_type), tmp_folder)

        dev_par, layers = load_device_parameters(session_path, device_parameters, run_mode = False)

        # check for new layers in the cmd_pars
        newlayers = []
        idx_layers = []
        for cmd_par in cmd_pars:
            if cmd_par['par'].startswith('l') and cmd_par['par'][1:].isdigit():
                newlayers.append(cmd_par)
                idx_layers.append(int(cmd_par['par'][1:]))
        
        # update layer files with cmd_pars
        if len(newlayers) > 0:
            for layer in layers:
                for cmd_par in cmd_pars:
                    if cmd_par['par'] == layer[1]:
                        layer[2] = cmd_par['val']
                        break
        
        # check if we need to add new layers
        newlayers = [x for _, x in sorted(zip(idx_layers, newlayers))]
        idx_layers = sorted(idx_layers)
        if len(idx_layers) > 0:
            if max(idx_layers) > len(layers)-1: # means we need to add new layers
                # check that all the layers between len(layers)-& and max(idx_layers) are in the idx_layers list
                for i in range(len(layers),max(idx_layers)+1):
                    if i not in idx_layers:
                        raise ValueError('Missing layer definition for layer '+str(i)+' in the command parameters list.')
                # add the new layers
                for i, layer in zip(idx_layers, newlayers):
                    layers.append(['par','l'+str(i),layer['val']],'parameter file for layer '+str(i))
            
                # update dev_par layer files with cmd_pars
                dev_par_keys = list(dev_par.keys())
                for i, section in enumerate(dev_par[dev_par_keys[0]]):
                    if section[0].lower() == 'layers':
                        # update section with layers[1:]
                        dev_par[dev_par_keys[0]][i][1:] = layers[1:]                  


        # move all layers to the temporary folder
        for layer in layers:
            if os.name == 'nt':
                layer_path = convert_to_long_path(os.path.join(session_path, layer[2]))
                shutil.copy(layer_path, tmp_folder)
            else:
                shutil.copy(os.path.join(session_path, layer[2]), tmp_folder)


        res = store_file_names(dev_par, sim_type, device_parameters, layers, run_mode = False)
        layer_files = res[0]
        optical_files = res[1]
        # make absolute paths
        layer_files = [os.path.abspath(os.path.join(session_path, f)) for f in layer_files]
        optical_files = [os.path.abspath(os.path.join(session_path, f)) for f in optical_files]
        optical_file_basenames = [os.path.basename(f) for f in optical_files]   
        traps_int_files = res[2]
        traps_bulk_files = res[3]
        traps_int_files = [os.path.abspath(os.path.join(session_path, f)) for f in traps_int_files]
        traps_bulk_files = [os.path.abspath(os.path.join(session_path, f)) for f in traps_bulk_files]
        traps_int_file_basenames = [os.path.basename(f) for f in traps_int_files]
        traps_bulk_file_basenames = [os.path.basename(f) for f in traps_bulk_files]

        ExpJV_file = None
        tVGFile = None
        tJFile = None
        JVFile = None
        scParsFile = None
        if sim_type == 'simss':
            if res[4].lower() == 'None'.lower():
                ExpJV_file = None
            else:
                ExpJV_file = res[4]
            JVFile = res[7]
            scParsFile = res[8]
        else:
            tVGFile = res[4]
            tJFile = res[7]

        varFile = res[5]
        logFile = res[6]


        # Copy the files to the temporary folder
        for file in layer_files + optical_files + traps_int_files + traps_bulk_files + [ExpJV_file] + [tVGFile]:
            if file is not None and os.path.isfile(file):
                if os.name == 'nt':
                    file_path = convert_to_long_path(file)
                    shutil.copy(file_path, tmp_folder)
                else:
                    shutil.copy(file, tmp_folder)

            # update temp folder files with basename
            if file in layer_files:
                make_basename_input_files(os.path.join(tmp_folder, os.path.basename(file)))       

        # move files from cmd_pars to the temporary folder and overwrite if necessary
        input_files = get_inputFile_from_cmd_pars(sim_type, cmd_pars)
        input_files = [os.path.abspath(os.path.join(session_path, f['val'])) for f in input_files]
        input_files_basenames = [os.path.basename(f) for f in input_files]
        for file in input_files:
            # if file already in the tmp_folder, remove it and copy the new one
            if os.path.isfile(os.path.join(tmp_folder, os.path.basename(file))):
                os.remove(os.path.join(tmp_folder, os.path.basename(file)))
            shutil.copy(os.path.join(session_path, file), tmp_folder)

        # set basename for device parameters
        make_basename_input_files(os.path.join(tmp_folder, os.path.basename(device_parameters)))


        # check the input files for the new layers
        # for layer in newlayers:
        #     input_files2 = get_inputFile_from_layer(layer,session_path)
        #     for file in input_files2:
        #         # if file does not exist in the tmp_folder, copy it. Otherwise, copy it if it is not in the input_files list
        #         if not os.path.isfile(os.path.join(tmp_folder, os.path.basename(file))):
        #             shutil.copy(os.path.join(session_path, file), tmp_folder)
        #         else:
        #             if not os.path.basename(file) in input_files_basenames:
        #                 shutil.copy(os.path.join(session_path, file), tmp_folder)

        lock.release()

        # Construct the command to run the executable
        cmd_pars = make_basename_file_cmd_pars(cmd_pars)
        cmd_line = construct_cmd(sim_type, cmd_pars)

        # Run the simulation
        result = run(cmd_line, cwd=tmp_folder, stdout=PIPE, check=False, shell=True)

        # Check the results of the process using the returncodes and console output
        if result.returncode != 0 and result.returncode != 95 and result.returncode != 3:
            # SIMsalabim raised an error, stop the program and return the error message on the UI.
            startMessage = False
            message = ''
            result_decode = result.stdout.decode('utf-8')

            if result.returncode >= 100:
                # A fatal (numerical) error occurred. Return errorcode and a standard error message.
                message = fatal_error_message(result.returncode)
            else:
                # Simsalabim raised an error. Read the console output for the details / error messaging. 
                # All linetypes after 'Program will be terminated, press enter.' are considered part of the error message
                for line_console in result_decode.split('\n'):
                    if startMessage is True:
                        # The actual error message. Since the error message can be multi-line, append each line.
                        message = message + line_console + '\n'
                    if 'Program will be terminated.' in line_console:
                        # Last 'regular' line of the console output. The next line is from the error message.
                        startMessage = True

            # Show the message as an error on the screen. Do not continue to the simulation results page.
            message = 'Simulation raised an error with Errorcode: ' + str(result.returncode) + '\n\n' + message
        else:
            # SIMsalabim simulation succeeded
            if result.returncode == 95:
                # In case of errorcode 95, failures during the simulations were encountered but the simulation did not halt. Show 'error' messages on the UI.
                startMessage = False
                message = ''
                result_decode = result.stdout.decode('utf-8')
                for line_console in result_decode.split('\n'):
                    if startMessage is True:
                        # The actual error message. Since the error message can be multi-line, append each line.
                        message = message + line_console + '\n'
                    if 'Program will be terminated.' in line_console:
                        # Last 'regular' line of the console output. The next line is from the error message.
                        startMessage = True

                # Show the message as a success on the screen
                message = 'Simulation completed but raised errorcode: ' + str(result.returncode) + '\n\n' + 'The simulation finished but at least 1 point did not converge. \n\n' + message
            elif result.returncode == 3:
                # Special case, should not occur in the web version.
                # When the program exits as a success but no simulation has been run, e.g. in the case of the autotidy functionality. 
                message = 'Action completed'
            else:
                # Simulation completed as expected.
                message = 'Simulation complete. Output can be found in the Simulation results.'

        # Move output files to the original folder
        lock.acquire()
        if sim_type.lower() == 'simss':
            if os.path.isfile(os.path.join(tmp_folder, varFile)):
                varFile_path = convert_to_long_path(os.path.join(tmp_folder, varFile))
                newFile_path = convert_to_long_path(os.path.join(session_path, varFile))
                shutil.move(varFile_path, newFile_path)
            if os.path.isfile(os.path.join(tmp_folder, logFile)):
                logFile_path = convert_to_long_path(os.path.join(tmp_folder, logFile))
                newLog_path = convert_to_long_path(os.path.join(session_path, logFile))
                shutil.move(logFile_path, newLog_path)
            if os.path.isfile(os.path.join(tmp_folder, JVFile)):
                JVFile_path = convert_to_long_path(os.path.join(tmp_folder, JVFile))
                shutil.move(JVFile_path, session_path)
            if os.path.isfile(os.path.join(tmp_folder, scParsFile)):
                scParsFile_path = convert_to_long_path(os.path.join(tmp_folder, scParsFile))
                newScPars_path = convert_to_long_path(os.path.join(session_path, scParsFile))
                shutil.move(scParsFile_path, newScPars_path)
        elif sim_type.lower() == 'zimt':
            if os.path.isfile(os.path.join(tmp_folder, varFile)):
                varFile_path = convert_to_long_path(os.path.join(tmp_folder, varFile))
                newFile_path = convert_to_long_path(os.path.join(session_path, varFile))
                shutil.move(varFile_path, newFile_path)
            if os.path.isfile(os.path.join(tmp_folder, logFile)):
                logFile_path = convert_to_long_path(os.path.join(tmp_folder, logFile))
                newLog_path = convert_to_long_path(os.path.join(session_path, logFile))
                shutil.move(logFile_path, newLog_path)
            if os.path.isfile(os.path.join(tmp_folder, tVGFile)):
                tVGFile_path = convert_to_long_path(os.path.join(tmp_folder, tVGFile))
                newTVG_path = convert_to_long_path(os.path.join(session_path, tVGFile))
                shutil.move(tVGFile_path, newTVG_path)
            if os.path.isfile(os.path.join(tmp_folder, tJFile)):
                tJFile_path = convert_to_long_path(os.path.join(tmp_folder, tJFile))
                newTJ_path = convert_to_long_path(os.path.join(session_path, tJFile))
                shutil.move(tJFile_path, newTJ_path)
        lock.release()

        # Release semaphore
        semaphore.release()

        # Print output if verbose
        if verbose:
            print(result.stdout)

        # Notify the queue that the task is done
        q.task_done()
    
    return result, message

def worker_linux(q,lock,semaphore,verbose=False):
    """Worker function that runs the simulation in a temporary folder and moves the output files to the original folder. 

    Parameters
    ----------
    q : queue
        Queue to communicate with the main thread
    lock : threading.Lock
        Lock to prevent multiple threads from writing to the same file at the same time
    semaphore : threading.Semaphore
        Semaphore to limit the number of parallel jobs
    verbose : bool
        If True, print the output of the simulation to the console

    """
    while True:
        # Get task from the queue
        task = q.get()
        if task is None:
            break
        
        # Unpack task
        sim_type, cmd_pars, session_path = task

        # Acquire semaphore
        semaphore.acquire()

        result, message = run_simulation(sim_type, cmd_pars, session_path, run_mode = False, verbose=verbose)

        # Release semaphore
        semaphore.release()

        # Print output if verbose
        if verbose:
            print(result.stdout)

        # Notify the queue that the task is done
        q.task_done()

        return result, message

def run_simulation_multithreaded_linux(sim_type,cmd_pars_list,session_path,max_jobs=max(1,os.cpu_count()-1),verbose=False):
    """Runs simulations in parallel on max_jobs number of threads.  
    This procedure should work on Windows and Linux but it is not as efficient as run_parallel_simu on Linux.
    Yet, it is the only way to run simulations in parallel on Windows in a thread safe way and making sure that two thread do not try to write to the same file at the same time.
    This is achieved by running the simulation in a temporary folder with a copy of all the necessary input files and then moving the output files to the original folder.


    Parameters
    ----------
    sim_type : string
        Which type of simulation to run: simss or zimt
    cmd_pars_list : List
        List of list with parameters to add to the simss/zimt cmd line. Each parameter is a dict with par,val keys. 
        Note: when relevant the first entry must be the deviceparameters file with a key: dev_par_file
    session_path : string
        File path of the simss or zimt executable 
    max_jobs : int
        Maximum number of parallel jobs to run. Default is the number of CPU cores - 1
    verbose : bool
        If True, print the output of the simulation to the console
    
    Returns
    -------
    List
        List of CompletedProcess objects with returncode and console output of the simulation

    """    
 
    lock = threading.Lock()    # Create a lock   
    semaphore = threading.Semaphore(max_jobs)  # Create a semaphore

    # Create a queue to communicate with the worker threads
    q = queue.Queue()

    # Start worker threads
    threads = []
    for i in range(len(cmd_pars_list)):
        t = CustomThread(target=partial(worker_linux,q=q,lock=lock,semaphore=semaphore,verbose=verbose))
        t.start()
        threads.append(t)

    # Add tasks to the queue
    for code_name, cmd_pars in zip([sim_type] * len(cmd_pars_list), cmd_pars_list):
        q.put((code_name, cmd_pars, session_path))

    # Wait for all tasks to be finished
    q.join()

    # Stop workers
    for i in range(len(cmd_pars_list)):
        q.put(None)

    result_list = []
    for t in threads:
        result_list.append(t.join())
    print(result_list)
    return_code_list = [res[0] for res in result_list]
    # message_list = [parallel_error_message(res) for res in return_code_list]
    message_list = []
    for res in return_code_list:
        if res == 91:
            # look for the logfile to get more information about the error
            for cmd_pars in cmd_pars_list:
                for c in cmd_pars:
                    if c['par'] == 'logFile':
                        logFile = c['val']
                        break
            if os.path.isfile(os.path.join(session_path,logFile)):
                # find line with 'Program will be terminated.' and store the next lines as the error message
                startMessage = False
                message = ''
                with open(os.path.join(session_path,logFile),'r') as f:
                    for line_console in f:
                        if startMessage is True:
                            # The actual error message. Since the error message can be multi-line, append each line.
                            message = message + line_console + '\n'
                        if 'Program will be terminated.' in line_console:
                            # Last 'regular' line of the console output. The next line is from the error message.
                            startMessage = True
                if message == '':
                    message = 'Simulation raised an error with Errorcode: ' + str(res) + '\n\n' + parallel_error_message(res)
                message_list.append(message)
            else:
                message_list.append('Simulation raised an error with Errorcode: ' + str(res) + '\n\n' + parallel_error_message(res))
        else:
            message_list.append(parallel_error_message(res))
    if not all(val in [0, 95, 3] for val in return_code_list):
        # check if only one error code different from 0, 95 or 3
        if len(set([val for val in return_code_list if val not in [0, 95, 3]])) == 1:
            # if so, return that error code and the corresponding message
            idx = return_code_list.index([val for val in return_code_list if val not in [0, 95, 3]][0])
            result = return_code_list[idx]
            message = message_list[idx]
            result_list = [(result, message)]
        else:
            print(f"Multiple different errors occurred during the parallel simulations: {set([val for val in return_code_list if val not in [0, 95, 3]])}. Returning error code 666.")
            result = 666
    elif all(val in [0, 95] for val in return_code_list):
        result = 95
    elif all(val in [0, 3] for val in return_code_list):
        result = 3
    else:
        result = 0
    return result, message_list, return_code_list
    # return result_list

# Custom thread class to return the result of the thread
class CustomThread(Thread):
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs={}, Verbose=None):
        Thread.__init__(self, group, target, name, args, kwargs)
        self._return = None
 
    def run(self):
        if self._target is not None:
            self._return = self._target(*self._args, **self._kwargs)
             
    def join(self, *args):
        Thread.join(self, *args)
        return self._return
















"""Perform impedance simulations"""
######### Package Imports #########################################################################

import os, sys
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import math
import scipy.integrate
# import pySIMsalabim
## Import pySIMsalabim, if not successful, add the parent directory to the system path
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.utils import general as utils_gen
from pySIMsalabim.plots import plot_functions as utils_plot
from pySIMsalabim.utils.utils import *
from pySIMsalabim.utils.device_parameters import *

######### Functions #################################################################################

def create_tVG_impedance(V_0, del_V, G_frac, tVG_name, session_path, f_min, f_max, ini_timeFactor, timeFactor):
    """Create a tVG file for impedance experiments. 

    Parameters
    ----------
    V_0 : float 
        Voltage at t=0
    del_V : float
        Voltage step that is applied after t=0
    G_frac : float
        Fractional light intensity
    tVG_name : string
        Name of the tVG file
    session_path : string
        Path of the simulation folder for this session
    f_min : float
        Minimum frequency
    f_max : float
        Maximum frequency
    ini_timeFactor : float
        Constant defining the size of the initial timestep
    timeFactor : float
        Exponential increase of the timestep, to reduce the amount of timepoints necessary. Use values close to 1.

    Returns
    -------
    string
        A message to indicate the result of the process
    """

    time = 0
    del_t = ini_timeFactor/f_max

    # Starting line of the tVG file: header + datapoints at time=0. Set the correct header
    tVG_lines = 't Vext G_frac\n' + f'{time:.3e} {V_0} {G_frac:.3e}\n'


    # Make the other lines in the tVG file
    while time < 1/f_min: #max time: 1/f_min is enough!
        time = time+del_t
        tVG_lines += f'{time:.3e} {V_0+del_V} {G_frac:.3e}\n'
        del_t = del_t * timeFactor # At first, we keep delt constant to its minimum values

    # Write the tVG lines to the tVG file
    with open(os.path.join(session_path,tVG_name), 'w') as file:
        file.write(tVG_lines)

    # tVG file is created, message a success
    msg = 'Success'
    retval = 0

    return retval, msg

def create_tVG_SS(V_0, G_frac, tVG_name, session_path):
    """ Creates the tVG file for the steady state simulation with only t 0 and V_0

    Parameters
    ----------
    V_0 : float 
        Voltage at t=0
    del_V : float
        Voltage step that is applied after t=0
    G_frac : float
        Fractional light intensity
    tVG_name : string
        Name of the tVG file
    session_path : string
        Path of the simulation folder for this session

    Returns
    -------
    string
        A message to indicate the result of the process

    
    """    

    # Starting line of the tVG file: header + datapoints at time=0. Set the correct header
    tVG_lines = 't Vext G_frac\n' + f'{0} {V_0} {G_frac:.3e}\n'
   
    # Write the tVG lines to the tVG file
    with open(os.path.join(session_path,tVG_name), 'w') as file:
        file.write(tVG_lines)

    # tVG file is created, message a success
    msg = 'Success'
    retval = 0

    return retval, msg

def create_tVG_tolDens(V_0, del_V, ini_timeFactor, f_min, f_max, G, tVG_name, session_path):
    """ Creates the tVG file for the steady state simulation with only t 0 and V_0

    Parameters
    ----------
    V_0 : float 
        Voltage at t=0
    del_V : float
        Voltage step that is applied after t=0
    G_frac : float
        Fractional light intensity
    tVG_name : string
        Name of the tVG file
    session_path : string
        Path of the simulation folder for this session

    Returns
    -------
    string
        A message to indicate the result of the process
    """    

    del_t = ini_timeFactor/f_max # t_1 = 0 s + del_t

    tVG_lines = ('t\tVext\tG_frac\n'
                 f'0\t{V_0+del_V}\t{G:.3e}\n'
                 f'0\t{V_0}\t{G:.3e}\n'
                 f'{del_t}\t{V_0+del_V}\t{G:.3e}\n')

    # Write the tVG lines to the tVG file
    with open(os.path.join(session_path,tVG_name), 'w') as file:
        file.write(tVG_lines)

    # tVG file is created, message a success
    msg = 'Success'
    retval = 0

    return retval, msg

def calc_impedance_limit_time(I, errI, time, VStep, imax):
    """Fourier Decomposition formula which computes the impedance at frequency freq (Hz) and its complex error
    Based on S.E. Laux, IEEE Trans. Electron Dev. 32 (10), 2028 (1985), eq. 5a, 5b
    We integrate from 0 to time[imax] at frequency 1/time[imax]

    Parameters
    ----------
    I : np.array
        Array of currents
    errI : np.array
        Numerical error in calculated currents (output of ZimT)
    time : np.array
        Array with all time positions, from 0 to tmax
    VStep : float
        Voltage step
    imax : integer
        Index of the last timestep/first frequency for which the integrals are calculated

    Returns
    -------
    float
        Frequency belonging to Z(f)
    complex number
        Impedance at frequency f: Z(f)
    complex number
        Numerical error in calculated impedance Z(f)
    """

    freq=1/time[imax] #we obtain the frequency from the time array
    Iinf = I[imax] # I at infinite time, i.e. the last one we have.
	
    #prepare array for integrants:
    int1 = np.empty(imax)
    int2 = np.empty(imax)
    int3 = np.empty(imax)
    int4 = np.empty(imax)
	
    #now we use only part of the time array:
    timeLim = time[0:imax]
	
    for i in range(imax) :
        sinfac = math.sin(2*math.pi*freq*timeLim[i])
        cosfac = math.cos(2*math.pi*freq*timeLim[i])
        int1[i] = sinfac*(I[i] - Iinf)
        int2[i] = cosfac*(I[i] - Iinf)	
        int3[i] = sinfac*(I[i] + errI[i] - Iinf - errI[imax])
        int4[i] = cosfac*(I[i] + errI[i] - Iinf - errI[imax])	

    #now compute the conductance and capacitance:
    cond = (Iinf - I[0] + 2*math.pi*freq*scipy.integrate.trapezoid(int1, timeLim))/VStep
    cap = scipy.integrate.trapezoid(int2, timeLim)/VStep
    #convert to impedance:
    Z = 1/(cond + 2J*math.pi*freq*cap)
	
    #and again, but now with the error added to the current:	
    condErr = (Iinf + errI[imax] - I[0] - errI[0] + 2*math.pi*freq*scipy.integrate.trapezoid(int3, timeLim))/VStep
    capErr = scipy.integrate.trapezoid(int4, timeLim)/VStep
    #convert to impedance:
    Z2 = 1/(condErr + 2J*math.pi*freq*capErr)
    
    #error is the difference between Z and Z2:
    errZ = Z - Z2
    
    #now return complex impedance, its error and the corresponding frequency:	
    return freq, Z, errZ

def calc_impedance(data, del_V, isToPlot,session_path,zimt_device_parameters,Rseries=0,Rshunt=-1e3):
    """ Calculate the impedance over the frequency range
    
    Parameters
    ----------
    data : dataFrame
        Pandas dataFrame containing the time, voltage, current density and numerical error in the current density of the tj_file
    del_V : float
        Voltage step
    isToPlot : list
        List of array indices that will be used in the plotting

    Returns
    -------
    np.array
        Array of frequencies
    np.array
        Array of the real component of impedance
    np.array
        Array of the imaginary component of impedance
    np.array
        Array of complex error
    np.array
        Array of capacitance	
    np.array
        Array of conductance
    np.array
        Array of error in capacitance	
    np.array
        Array of error in conductance
    """
    # init the arrays for the impedance and its error:
    numFreqPoints = len(isToPlot)
    freq = np.empty(numFreqPoints)
    ReZ = np.empty(numFreqPoints)
    ImZ = np.empty(numFreqPoints)
    Z = [1 + 1J] * numFreqPoints
    errZ = [1 + 1J] * numFreqPoints
    C = np.empty(numFreqPoints)
    G = np.empty(numFreqPoints)
    errC = np.empty(numFreqPoints)
    errG = np.empty(numFreqPoints)

    for i in range(numFreqPoints):
        imax=isToPlot[i]
        freq[i], Z[i], errZ[i] = calc_impedance_limit_time(data['Jext'], data['errJ'], data['t'], del_V, imax)
        if Rseries > 0 and Rshunt > 0:
            # Correct the impedance for the series resistance
            Z[i] = Rseries + 1/(1/Z[i] + 1/Rshunt)
        elif Rseries > 0 and Rshunt < 0:
            # Correct the impedance for the series resistance
            Z[i] = Rseries + Z[i]
        elif Rshunt > 0 and Rseries <= 0:
            # Correct the impedance for the shunt resistance
            invZ = 1/Z[i]
            Z[i] = 1/(invZ + 1/Rshunt)

        invZ = 1/Z[i]
        
        # we are only interested in the absolute value of the real and imag components:
        ReZ[i] = Z[i].real
        ImZ[i] = Z[i].imag
        C[i] = 1/(2*math.pi*freq[i])*invZ.imag
        G[i] = invZ.real

        errC[i] = abs(1/(2*math.pi*freq[i])*(invZ.imag**2)*errZ[i].real)
        errG[i] = abs((invZ.real**2)*errZ[i].imag)
    
    return freq, ReZ, ImZ, errZ, C, G, errC, errG

def store_impedance_data(session_path, freq, ReZ, ImZ, errZ, C, G, errC, errG, output_file):
    """ Save the frequency, real & imaginary part of the impedance & impedance error in one file called freqZ.dat
    
    Parameters
    ----------
    session_path : string
        working directory for zimt
    freq : np.array
        Array of frequencies
    ReZ : np.array
        Array of the real component of impedance
    ImZ : np.array
        Array of the imaginary component of impedance
    errZ : np.array
        Array of complex error
    C : np.array
        Array of capacitance	
    G : np.array
        Array of conductance
    errC : np.array
        Array of error in capacitance	
    errG : np.array
        Array of error in conductance
    """

    with open(os.path.join(session_path,output_file), 'w') as file:
        file.write('freq ReZ ImZ ReErrZ ImErrZ C G errC errG' + '\n')
        for i in range(len(freq)):
            file.write(f'{freq[i]:.6e} {ReZ[i]:.6e} {ImZ[i]:.6e} {abs(errZ[i].real):.6e} {abs(errZ[i].imag):.6e} {C[i]:.6e} {G[i]:.6e} {errC[i]:.6e} {errG[i]:.6e}\n')

    # print('The data of the Impedance Spectroscopy graphs is written to ' + output_file)

def get_impedance(data, f_min, f_max, f_steps, del_V, session_path, output_file,zimt_device_parameters,Rseries=0,Rshunt=-1e3):
    """Calculate the impedance from the simulation result

    Parameters
    ----------
    data : DataFrame
        DataFrame with the simulation results (tj.dat) file
    f_min : float
        Minimum frequency
    f_max : float
        Maximum frequency
    f_steps : float
        Frequency step
    del_V : float
        Voltage step
    session_path : string
        working directory for zimt
    output_file : string
        name of the file where the impedance data is stored

    Returns
    -------
    integer,string
        returns -1 (failed) or 1 (success), including a message
    """
    isToPlot, msg = get_integral_bounds(data, f_min, f_max, f_steps)

    if isToPlot != -1:
        # Integral bounds have been determined, continue to calculate the impedance
        freq, ReZ, ImZ, errZ, C, G, errC, errG = calc_impedance(data, del_V, isToPlot,session_path,zimt_device_parameters,Rseries,Rshunt)

        # Write impedance results to a file
        store_impedance_data(session_path, freq, ReZ, ImZ, errZ, C, G, errC, errG, output_file)

        msg = 'Success'
        return 0, msg
    else:
        # Failed to determine integral bounds, exit with the error message
        return -1, msg

def Bode_plot(session_path, output_file, xscale='log', yscale_1='linear', yscale_2='linear', plot_type = plt.errorbar):
    """ Plot the real and imaginary part of the impedance against frequency

    Parameters
    ----------
    session_path : string
        working directory for zimt
    xscale : string
        Scale of the x-axis. E.g linear or log
    yscale_ax1 : string
        Scale of the left y-axis. E.g linear or log
    yscale_ax2 : string
        Scale of the right y-axis. E.g linear or log
    """
    # Read the data from freqZ-file
    data = pd.read_csv(os.path.join(session_path,output_file), sep=r'\s+')

    # Flip the ImZ data to the first quadrant
    data["ImZ"] = data["ImZ"]*-1

    # Define the plot parameters, two y axis
    pars = {'ReZ' : 'Re Z [Ohm m$^2$]', 'ImZ' : '-Im Z [Ohm m$^2$]' }
    selected_1 = ['ReZ']
    selected_2 = ['ImZ']
    par_x = 'freq'
    xlabel = 'frequency [Hz]'
    ylabel_1 = 'Re Z [Ohm m$^2$]'
    ylabel_2 = '-Im Z [Ohm m$^2$]'
    title = 'Bode plot'

    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()
    utils_plot.plot_result_twinx(data, pars, selected_1, selected_2, par_x, xlabel, ylabel_1, ylabel_2, xscale, yscale_1, yscale_2, title,ax1,ax2, 
                                        plot_type, y_error_1 = data['ReErrZ'], y_error_2 = data['ImErrZ'])
    plt.show()

def Nyquist_plot(session_path, output_file, xscale='linear', yscale='linear', plot_type = plt.errorbar):
    """ Plot the real and imaginary part of the impedance against each other

    Parameters
    ----------
    session_path : string
        working directory for zimt
    xscale : string
        Scale of the x-axis. E.g linear or log
    yscale : string
        Scale of the y-axis. E.g linear or log
    """
    # Read the data from freqZ-file
    data = pd.read_csv(os.path.join(session_path,output_file), sep=r'\s+')

    # Flip the ImZ data to the first quadrant
    data["ImZ"] = data["ImZ"]*-1

    fig, ax = plt.subplots()
    pars = {'ImZ' : '-Im Z [Ohm m$^2$]'}
    par_x = 'ReZ'
    xlabel = 'Re Z [Ohm m$^2$]'
    ylabel = '-Im Z [Ohm m$^2$]'
    title = 'Nyquist plot'

    # Plot the nyquist plot with or without errorbars
    if plot_type == plt.errorbar:
        ax = utils_plot.plot_result(data, pars, list(pars.keys()), par_x, xlabel, ylabel, xscale, yscale, title, ax, plot_type, 
                                            data['ReErrZ'], data['ImErrZ'], legend=False)
    else:
        ax = utils_plot.plot_result(data, pars, list(pars.keys()), par_x, xlabel, ylabel, xscale, yscale, title, ax, plot_type, legend=False)

    plt.show()


def Capacitance_plot(session_path, output_file, xscale='log', yscale='linear'):
    """ Plot the capacitance against frequency

    Parameters
    ----------
    session_path : string
        working directory for zimt
    xscale : string
        Scale of the x-axis. E.g linear or log
    yscale : string
        Scale of the y-axis. E.g linear or log
    """
    # Read the data from freqZ-file
    data = pd.read_csv(os.path.join(session_path,output_file), sep=r'\s+')

    # Flip the ImZ data to the first quadrant
    data["C"] = data["C"]

    fig, ax = plt.subplots()
    pars = {'C' : 'C [F m$^{-2}$]'}
    par_x = 'freq'
    xlabel = 'frequency [Hz]'
    ylabel = 'C [F m$^{-2}$]'
    title = 'Capacitance plot'

    ax = utils_plot.plot_result(data, pars, list(pars.keys()), par_x, xlabel, ylabel, xscale, yscale, title, ax, plt.errorbar, y_error=data['errC'], legend=False)

    plt.show()

def plot_impedance(session_path, output_file='freqZ.dat'):
    """Make a Bode and Nyquist plot of the impedance

    Parameters
    ----------
    session_path : string
        working directory for zimt
    """
    # Bode plot
    Bode_plot(session_path,output_file)

    #Nyquist plot
    Nyquist_plot(session_path,output_file)

    # Capacitance plot
    Capacitance_plot(session_path,output_file)


def get_tolDens(zimt_device_parameters, session_path, f_min, f_max, V_0, G_frac, del_V, run_mode, tVG_name, tj_name, varFile, ini_timeFactor, dum_str, cmd_pars):
    """
    Calculate the tolerance of the density solver, to ensure a reliable impedance spectrum

    Parameters
    ----------
    zimt_device_parameters : string
        Name of the zimt device parameters file
    session_path : string
        Working directory for zimt
    f_min : float
        Minimum frequency
    f_max : float
        Maximum frequency
    V_0 : float
        Voltage at t=0
    G_frac : float
        Fractional light intensity
    del_V : float
        Voltage step
    run_mode : bool
        Indicate whether the script is in 'web' mode (True) or standalone mode (False). Used to control the console output
    tVG_name : string
        Name of the tVG file
    tj_name : string
        Name of the tj file
    varFile : string
        Name of the var file
    ini_timeFactor : float
        Constant defining the size of the initial timestep
    dum_str : string
        dummy string with UUID string to append to the file names
    cmd_pars : list
        List of dictionaries with the command line parameters
    
    Returns
    -------
    integer
        Return code of the simulation, 0 if successful, -1 if one of the steps failed else the return code of the simulation
    string
        Return message to display on the UI in case of failure
    float
        Tolerance of the density solver. If failed, returns None
    """

    # Determine J(t=0), J(t=∞), and J_dis to get the maximum allowed tolerance of the density solver to retrieve a
    # reliable impedance spectrum, where tolDens = J(t=∞) - J(t=0) / J_displacement
    result, message = create_tVG_tolDens(V_0, del_V, ini_timeFactor, f_min, f_max, G_frac, tVG_name, session_path)
    
    if result == 0:
        # In order for zimt to converge, set absolute tolerance of Poisson solver small enough
        tolPois = 10**(math.floor(math.log10(abs(del_V)))-5)
        
        tolDens_test = str(1e-10)
        # Define mandatory options for ZimT to run well with impedance:
        tolDens_args = [{'par':'dev_par_file','val':zimt_device_parameters},
                             {'par':'tVGFile','val':tVG_name},
                             {'par':'tolPois','val':str(tolPois)},
                             {'par':'tolDens','val':tolDens_test},
                             {'par':'limitDigits','val':'0'},
                             {'par':'currDiffInt','val':'2'},
                             {'par':'tJFile','val':tj_name},
                             {'par':'varFile','val':varFile},
                             {'par':'logFile','val':'log'+dum_str+'.txt'}]
        
        if cmd_pars is not None:
            tolDens_args = update_cmd_pars(tolDens_args, cmd_pars)
        
        result, message = utils_gen.run_simulation('zimt', tolDens_args, session_path, run_mode)

        if result == 0 or result == 95:
            data = read_tj_file(session_path, tj_file_name=tj_name)
            try:
                J_0 = data['Jext'][1]
                J_inf = data['Jext'][0]
                J_dis = abs(data['Jext'][2] - J_0)
            except KeyError as key:
                if key == 0:
                    message = f"J_inf does not exist in the tolDens tJ-file"
                elif key == 1:
                    message = f"J_0 does not exist in the tolDens tJ-file"
                elif key == 2:
                    message = f"J_spike does not exist in the tolDens tJ-file"
                return -1, message, None
            
            # tolDens cannot be larger than 1E-6
            tolDens = min(abs(J_inf - J_0) / J_dis * 1e-4, 1e-6)

            # tolDens cannot be smaller than 1E-12
            tolDens = max(tolDens, 1e-12)
        else:
            return result, message, None
        
        # Remove the tVG and tJ files as they are not needed anymore
        os.remove(os.path.join(session_path,tVG_name))
        os.remove(os.path.join(session_path,tj_name))
        
    else:
        message = "Computing tolDens was unsuccesful"
        return -1, message, None

    return 0, '', tolDens

def run_impedance_simu(zimt_device_parameters, session_path, f_min, f_max, f_steps, V_0, G_frac = 1, del_V = 0.01, run_mode = False, tVG_name='tVG.txt', output_file = 'freqZ.dat', tj_name = 'tj.dat', varFile ='none', ini_timeFactor=1e-3, timeFactor=1.02, **kwargs):
    """Create a tVG file and run ZimT with impedance device parameters

    Parameters
    ----------
    zimt_device_parameters : string
        Name of the zimt device parameters file
    session_path : string
        Working directory for zimt
    f_min : float
        Minimum frequency
    f_max : float
        Maximum frequency
    f_steps : float
        Frequency step
    V_0 : float 
        Voltage at t=0
    G_frac : float, optional
        Fractional light intensity, by default 1
    del_V : float, optional
        Voltage step, by default 0.01
    run_mode : bool, optional
        Indicate whether the script is in 'web' mode (True) or standalone mode (False). Used to control the console output, by default False  
    tVG_name : string, optional
        Name of the tVG file, by default tVG.txt
    output_file : string, optional
        Name of the file where the impedance data is stored, by default freqZ.dat
    tj_name : string, optional
        Name of the tj file where the impedance data is stored, by default tj.dat
    varFile : string, optional
        Name of the var file, by default 'none'
    ini_timeFactor : float, optional
        Constant defining the size of the initial timestep, by default 1e-3
    timeFactor : float, optional
        Exponential increase of the timestep, to reduce the amount of timepoints necessary. Use values close to 1., by default 1.02
        
    Returns
    -------
    CompletedProcess
        Output object of with returncode and console output of the simulation
    string
        Return message to display on the UI, for both success and failed
    """
    verbose = kwargs.get('verbose', False) # Check if the user wants to see the console output
    UUID = kwargs.get('UUID', '') # Check if the user wants to add a UUID to the tj file name
    cmd_pars = kwargs.get('cmd_pars', None) # Check if the user wants to add additional command line parameters
    # Check if the user wants to force the use of thread safe mode, necessary for Windows with parallel simulations
    if os.name == 'nt':
        threadsafe = kwargs.get('threadsafe', True) # Check if the user wants to force the use of threads instead of processes
    else:
        threadsafe = kwargs.get('threadsafe', False) # Check if the user wants to force the use of threads instead of processes
    
    turnoff_autoTidy = kwargs.get('turnoff_autoTidy', None) # Check if the user wants to turn off the autoTidy function in SIMsalabim
    if turnoff_autoTidy is None: 
        if not threadsafe:
            turnoff_autoTidy = True
        else:
            turnoff_autoTidy = False

    # Update the file names with the UUID
    if UUID != '':
        dum_str = f'_{UUID}'
    else:
        dum_str = ''

    # Update the filenames with the UUID
    tj_name = os.path.join(session_path, tj_name)
    output_file = os.path.join(session_path, output_file)
    tVG_name = os.path.join(session_path, tVG_name)
    if UUID != '':
        tj_file_name_base, tj_file_name_ext = os.path.splitext(tj_name)
        tj_name = tj_file_name_base + dum_str + tj_file_name_ext 
        tVG_name_base, tVG_name_ext = os.path.splitext(tVG_name)
        tVG_name = tVG_name_base + dum_str + tVG_name_ext
        output_file_base, output_file_ext = os.path.splitext(output_file)
        output_file = output_file_base + dum_str + output_file_ext
        if varFile != 'none':
            var_file_base, var_file_ext = os.path.splitext(varFile)
            varFile = var_file_base + dum_str + var_file_ext
            varFile = os.path.join(session_path,varFile)
    # varFile = 'none' # we don't use a var file for the hysteresis JV simulation

    ##############################################################################
    # If the voltage is set to Voc, firstly compute its value
    if V_0 == 'oc':
        # Create tVG
        result, message = create_tVG_SS(V_0, G_frac, tVG_name, session_path)
        
        # Check if tVG file is created
        if result == 0:
            # In order for zimt to converge, set absolute tolerance of Poisson solver small enough
            tolPois = 10**(math.floor(math.log10(abs(del_V)))-4)

            # Define mandatory options for ZimT to run well with impedance:
            Impedance_SS_args = [{'par':'dev_par_file','val':zimt_device_parameters},
                                {'par':'tVGFile','val':tVG_name},
                                {'par':'tolPois','val':str(tolPois)},
                                {'par':'limitDigits','val':'0'},
                                {'par':'currDiffInt','val':'2'},
                                {'par':'tJFile','val':tj_name},
                                {'par':'varFile','val':varFile},
                                {'par':'logFile','val':'log'+dum_str+'.txt'}
                                ]
            
            if turnoff_autoTidy:
                Impedance_SS_args.append({'par':'autoTidy','val':'0'})

            if cmd_pars is not None:
                Impedance_SS_args = update_cmd_pars(Impedance_SS_args, cmd_pars)
            
            if threadsafe:
                result, message = utils_gen.run_simulation_filesafe('zimt', Impedance_SS_args, session_path, run_mode, verbose=verbose)
            else:
                result, message = utils_gen.run_simulation('zimt', Impedance_SS_args, session_path, run_mode, verbose=verbose)
    
            if result == 0 or result == 95:
                data = read_tj_file(session_path, tj_file_name=tj_name)
                
                V_0 = data['Vext'][0]
            else:
                message = "Computing the value of Voc led to the following error: " + message
                return result, message
    else:
        V_0 = float(V_0)

    ##############################################################################
    # The simulations with Rseries and Rshunt often do not converge, so we first run a steady state simulation to get the internal voltage and then run the impedance simulation with Rseries = 0 and Rshunt = -Rshunt. We will correct the impedance afterwards. This is a workaround to improve the convergence of the impedance simulation that should remain accurate to estimate the impedance.
    # Default values for Rseries and Rshunt
    Rseries = 0
    Rshunt = -1 # Negative values are used for Rshunt in SIMsalabim to indicate infinite Rshunt

    # Do the steady state simulation to calculate the internal voltage in case of series resistance
    # Create tVG
    result, message = create_tVG_SS(V_0, G_frac, tVG_name, session_path)

    # Get the device parameters and Rseries and Rshunt
    dev_val,layers_dum = load_device_parameters(session_path, zimt_device_parameters)
    for i in dev_val[zimt_device_parameters]:
        if i[0] == 'Contacts':
            contacts = i
            break

    for i in contacts[1:]:
        if i[1] == 'R_series':
            Rseries = float(i[2])
        elif i[1] == 'R_shunt':
            Rshunt = float(i[2])

    # Check if R_series and R_shunt are defined in cmd_pars
    idx_Rseries, idx_Rshunt = None, None
    if cmd_pars is not None:
        for idx, i in enumerate(cmd_pars):
            if i['par'] == 'R_series':
                Rseries = float(i['val'])
                idx_Rseries = idx
            elif i['par'] == 'R_shunt':
                Rshunt = float(i['val'])
                idx_Rshunt = idx
     
    if Rseries > 0:
        # Check if tVG file is created
        if result == 0:
            # In order for zimt to converge, set absolute tolerance of Poisson solver small enough
            tolPois = 10**(math.floor(math.log10(abs(del_V)))-4)

            # Define mandatory options for ZimT to run well with impedance:
            Impedance_SS_args = [{'par':'dev_par_file','val':zimt_device_parameters},
                                {'par':'tVGFile','val':tVG_name},
                                {'par':'tolPois','val':str(tolPois)},
                                {'par':'limitDigits','val':'0'},
                                {'par':'currDiffInt','val':'2'},
                                {'par':'tJFile','val':tj_name},
                                {'par':'varFile','val':varFile},
                                {'par':'logFile','val':'log'+dum_str+'.txt'}
                                ]
            
            if turnoff_autoTidy:
                Impedance_SS_args.append({'par':'autoTidy','val':'0'})

            if cmd_pars is not None:
                Impedance_SS_args = update_cmd_pars(Impedance_SS_args, cmd_pars)
            
            if threadsafe:
                result, message = utils_gen.run_simulation_filesafe('zimt', Impedance_SS_args, session_path, run_mode, verbose=verbose)
            else:
                result, message = utils_gen.run_simulation('zimt', Impedance_SS_args, session_path, run_mode, verbose=verbose)
    
            if result == 0 or result == 95:
                data = read_tj_file(session_path, tj_file_name=tj_name)

                Vext = data['Vext'][0]
                Jext = data['Jext'][0]

                Vint = Vext - Jext*Rseries
                V_0 = Vint # we need to shift the voltage to the internal voltage to account for the series resistance

            else:
                return result, message
        else:
            return result, message

    # remove the Rseries and Rshunt from cmd_pars
    if cmd_pars is not None:
        cmd_pars = [dictionary for dictionary in cmd_pars if dictionary['par'] not in ('R_series', 'R_shunt')]

    # Calculate the tolerance of the density solver
    result, message, tolDens = get_tolDens(zimt_device_parameters, session_path, f_min, f_max, V_0, G_frac, del_V, run_mode, tVG_name, tj_name, varFile, ini_timeFactor, dum_str, cmd_pars)

    if result != 0:
        # Failed to calculate the tolerance of the density solver, return the error message
        return result, message

    # Do the impedance simulation
    # Create tVG
    result, message = create_tVG_impedance(V_0, del_V, G_frac, tVG_name, session_path, f_min, f_max, ini_timeFactor, timeFactor)

    # Check if tVG file is created
    if result == 0:
        # In order for zimt to converge, set absolute tolerance of Poisson solver small enough
        tolPois = 10**(math.floor(math.log10(abs(del_V)))-4)

        # Define mandatory options for ZimT to run well with impedance:
        Impedance_args = [{'par':'dev_par_file','val':zimt_device_parameters},
                             {'par':'tVGFile','val':tVG_name},
                             {'par':'tolPois','val':str(tolPois)},
                             {'par':'tolDens','val':str(tolDens)},
                             {'par':'limitDigits','val':'0'},
                             {'par':'currDiffInt','val':'2'},
                             {'par':'tJFile','val':tj_name},
                             {'par':'varFile','val':varFile},
                             {'par':'logFile','val':'log'+dum_str+'.txt'},
                             # We remove Rseries and Rshunt as the simulation is either to converge that way, we will correct the impedance afterwards
                             {'par':'R_series','val':str(0)},
                             {'par':'R_shunt','val':str(-1)}]
        
        if turnoff_autoTidy:
            Impedance_args.append({'par':'autoTidy','val':'0'})
            
        if cmd_pars is not None:
            Impedance_args = update_cmd_pars(Impedance_args, cmd_pars)

        if threadsafe:
            result, message = utils_gen.run_simulation_filesafe('zimt', Impedance_args, session_path, run_mode, verbose=verbose)
        else:
            result, message = utils_gen.run_simulation('zimt', Impedance_args, session_path, run_mode, verbose=verbose)

        if result == 0 or result == 95:
            data = read_tj_file(session_path, tj_file_name=tj_name) 
            result, message = get_impedance(data, f_min, f_max, f_steps, del_V, session_path, output_file, zimt_device_parameters, Rseries, Rshunt)
            return result, message

        else:
            return result, message

    return result, message

## Running the function as a standalone script
if __name__ == "__main__":
    ## Manual Impedance input parameters. These are overwritten by command line arguments if provided
    f_min = 1e-2 # org 1e-2
    f_max = 1e6 # org 1e6
    f_steps = 20 # org 20
    V_0 = 0.6 # Float or 'oc' for the open-circuit voltage
    del_V = 1e-2 # org 1e-2
    G_frac = 1

    # Define folder and file paths
    session_path = os.path.join('../../','SIMsalabim','ZimT')

    zimt_device_parameters = 'simulation_setup.txt'

    tVGFile = 'tVG.txt'
    tJFile = 'tj.dat'
    output_name = 'freqZ.dat'
    varFile = 'none'
    
    # UUID = str(uuid.uuid4()) # Add a UUID to the simulation
    UUID = ''

    # Not user input
    ini_timeFactor = 1e-3 # Initial timestep factor, org 1e-3
    timeFactor = 1.02 # Increase in timestep every step to reduce the amount of datapoints necessary, use value close to 1 as this is best! Org 1.02
    run_mode = False # Show verbose output in console

    ############## Command line arguments  ##############
    ## Notes
    ## - The command line arguments are optional and can be provided in any order
    ## - Each command line argument must be provided in the format -par_name value
    ## - Possible arguments include all SIMsalabim parameters and Impedance specific parameters as listed before
    ## - Special arguments
    ##   - sp : string
    ##     - The session path, i.e. the working directory for the simulation
    ##   - zimt_device_parameters : string
    ##     - The name of the zimt simulation setup parameters file
    ##   - UUID : string
    ##     - An UUID to add to the simulation (output)
    
    cmd_pars_dict = {}
    cmd_pars = []

    # Check if any arguments are provided.
    if len(sys.argv) >= 2:
        # Each arguments should be in the format -par val, i.e. in pairs of two. Skip the first argument as this is the script name
        if not len(sys.argv[1:]) % 2 == 0:
            print('Error in command line parameters. Please provide arguments in the format -par_name value -par_name value ...')
            sys.exit(1)
        else:
            input_list = sys.argv[1:]
            # Loop over the input list and put pairs into a dictionary with the first argument as key and the second argument as value
            for i in range(0,len(input_list),2):
                # Check if the key already exists, if not, add it to the cmd_pars_dict
                if str(input_list[i][1:]) in cmd_pars_dict:
                    print(f'Duplicate parameter found in the command line parameters: {str(input_list[i][1:])}')
                    sys.exit(1)
                else:
                    cmd_pars_dict[str(input_list[i][1:])] = str(input_list[i+1])

    # Check and process specific keys/arguments that are not native SIMsalabim arguments
    # Handle the session_path/sp argument separately, as the other parameters depend on this
    if 'sp' in cmd_pars_dict:
        session_path = cmd_pars_dict['sp']
        # remove from cmd_pars
        cmd_pars_dict.pop('sp')

    # Convert the session_path to an actual path type
    session_path = os.path.abspath(session_path)

    # Define mappings for keys and variables
    key_action_map = {
        'zimt_device_parameters': lambda val: {'zimt_device_parameters': val},
        'f_min': lambda val: {'f_min': float(val)},
        'f_max': lambda val: {'f_max': float(val)},
        'f_steps': lambda val: {'f_steps': int(val)},
        'V_0': lambda val: {'V_0': val},
        'del_V': lambda val: {'del_V': float(val)},
        'G_frac': lambda val: {'G_frac': float(val)},
        'tVGFile': lambda val: {'tVGFile': val},
        'tJFile': lambda val: {'tJFile': val},
        'varFile': lambda val: {'varFile': val},        
        'output_name': lambda val: {'output_name': val},
        'UUID': lambda val: {'UUID': val},
    } 

    # Use exactly the same names as in SIMsalabim and as the Manual input parameters, 
    # if not than the tVG file name is not updated if putting e.g. "-tVGFile tVG_1.txt" in the command line in the terminal. 
    # Instead two tVG files names are defined one from the manual input parameters and one from the command line wherefore 
    # SIMsalabim gives the error "invalid input"

    for key in list(cmd_pars_dict.keys()):  # Use list to avoid modifying the dictionary while iterating
        if key in key_action_map:
            # Apply the corresponding action
            result = key_action_map[key](cmd_pars_dict[key])
            globals().update(result)  # Dynamically update global variables
            cmd_pars_dict.pop(key)

    # Handle remaining keys in `cmd_pars_dict` and add them to the cmd_pars list
    cmd_pars.extend({'par': key, 'val': value} for key, value in cmd_pars_dict.items())

    ## Run impedance spectroscopy
    result, message = run_impedance_simu(zimt_device_parameters, session_path, f_min, f_max, f_steps, V_0, G_frac, del_V, run_mode=run_mode, tVG_name=tVGFile,
                                         output_file=output_name, tj_name=tJFile, varFile=varFile, ini_timeFactor=ini_timeFactor, timeFactor=timeFactor, cmd_pars=cmd_pars, UUID=UUID)

    # Make the impedance plots
    calc_Voc_output_string = 'Computing the value of Voc led to the following error:'
    if result == 0 or (result == 95 and calc_Voc_output_string not in message):
        plot_impedance(session_path, os.path.basename(output_name))
    else:
        print(message)
        sys.exit(1)




"""Functions for general use"""
######### Package Imports #########################################################################

import os, subprocess, uuid, shutil, time
import pandas as pd
from subprocess import run, PIPE
from pySIMsalabim.utils.device_parameters import *

######### Function Definitions ####################################################################

def fatal_error_message(errorcode):
    """When a 'standard Pascal' fatal error occurs, add the standard error message

    Parameters
    ----------
    errorcode : int
        the error code
    """
    message = ''
    if errorcode == 106:
        message = 'Invalid numeric format: Reported when a non-numeric value is read from a text file'
    elif errorcode == 200:
        message = 'Division by zero: The application attempted to divide a number by zero.'
    elif errorcode == 201:
        message = 'Range check error.'
    elif errorcode == 202:
        message = 'Stack overflow error: This error is only reported when stack checking is enabled.'
    elif errorcode == 205:
        message = 'Floating point overflow.'
    elif errorcode == 206:
        message = 'Floating point underflow.'
    else:
        message = 'A fatal error occured.'
    return message

def error_message(errorcode):
    """When a 'standard Pascal' fatal error occurs, add the standard error message to be used with parallel which does not read the 

    Parameters
    ----------
    errorcode : int
        the error code

    Returns
    -------
    str
        the error message

    """
    message = ''
    if errorcode >= 90 and errorcode < 100:
        message = 'Error '+str(errorcode) +': '
        if errorcode == 90:
            message += 'Device parameter file corrupted.'
        elif errorcode == 91:
            message += 'Invalid input (physics, or voltage in tVGFile too large).'
        elif errorcode == 92:
            message += 'Invalid input from command line.'
        elif errorcode == 93:
            message += 'Numerical failure.'
        elif errorcode == 94:
            message += 'Failed to converge, halt (FailureMode = 0).'
        elif errorcode == 95:
            message += ' Failed to converge at least 1 point, not halt (FailureMode != 0).'
        elif errorcode == 96:
            message += 'Missing input file.'
        elif errorcode == 97:
            message += 'Runtime exceeds limit set by timeout.'
        elif errorcode == 99:
            message += 'Programming error (i.e. not due to the user!).'
    elif errorcode > 100:
        message = 'Fatal error '+str(errorcode) +': '
        if errorcode == 106:
            message += 'Invalid numeric format: Reported when a non-numeric value is read from a text file.'
        elif errorcode == 200:
            message += 'Division by zero: The application attempted to divide a number by zero.'
        elif errorcode == 201:
            message += 'Range check error.'
        elif errorcode == 202:
            message += 'Stack overflow error: This error is only reported when stack checking is enabled.'
        elif errorcode == 205:
            message += 'Floating point overflow.'
        elif errorcode == 206:
            message += 'Floating point underflow.'
        elif errorcode == 666:
            message += 'Multiple simulations failed with different error codes during parallel execution.'
        else:
            message = 'Unknown error code '+str(errorcode) + ' occurred.'
    elif errorcode == 3:
        message = 'Warning-like exit.'
    else:
        message = 'Unknown error code '+str(errorcode) + ' occurred.'

    return message

def construct_cmd(sim_type, cmd_pars):
    """Construct a single string to use as command to run a SIMsalabim executable

    Parameters
    ----------
    sim_type : string
        Which program to run: simss or zimt
    cmd_pars : List
        List with parameters to add to the simss/zimt cmd line. Each parameter is a dict with par,val keys. 
        Note: when relevant the first entry must be the deviceparameters file with a key: dev_par_file

    Returns
    -------
    string
        Constructed string to run as cmd
    """
    # Start with the executable name
    cmd_line = './' + sim_type

    # if system is 'Windows' use the .exe extension
    if os.name == 'nt':
        cmd_line = sim_type + '.exe'
    

    # Check whether a device parameters file has been defined
    for i in cmd_pars:
        # When specified, the device parameter file must be placed first, as is required by SIMsalabim
        if i['par'] == 'dev_par_file':
            if os.name == 'nt':
                # Add the device parameters file as " -dev_par_file "dev_par_file"
                args_single = ' "' + i['val']+ '"'
            else:
                args_single = ' ''' + i['val']+ ' '''
            cmd_line = cmd_line + args_single
            # After the dev_par_file key had been found once, stop the loop. If more than one dev_par_file is specified, the rest are ignored.
            break

    # Add the parameters
    for i in cmd_pars:
        if i['par'] != 'dev_par_file':
            # Add each parameter as " -par_name par_value"
            if os.name == 'nt': #add  "" to the parameter name if on Windows
                args_single = ' -"' + i['par'] + '" "' + i['val'] + '"'
            else:
                args_single = ' -' +i['par'] + ' ' + i['val']
            cmd_line = cmd_line + args_single
    
    return cmd_line

def run_simulation(sim_type, cmd_pars, session_path, run_mode = False, verbose = False):
    """Run the SIMsalabim simulation executable with the chosen device parameters. 
        Return the complete result object of the process accompanied by a message with information, 
        in case of both success and failure.

    Parameters
    ----------
    sim_type : string
        Which type of simulation to run: simss or zimt
    cmd_pars : List
        List with parameters to add to the simss/zimt cmd line. Each parameter is a dict with par,val keys. 
        Note: when relevant the first entry must be the deviceparameters file with a key: dev_par_file
    session_path : string
        File path of the simss or zimt executable 
    run_mode : boolean
        True if function is called as part of The Shell, False when called directly. 
        Prevents using streamlit components outside of The Shell.
    verbose : boolean
        True if the console output of the simulation should be printed to the console

    Returns
    -------
    int
        Return code of the simulation process, 0 for success, other values for errors.
    string
        Return message to display on the UI, for both success and failed
    """
    # Construct the command to run the executable
    cmd_line = construct_cmd(sim_type, cmd_pars)
    
    if run_mode:
        # Run the simulation in a shell environment 
        if os.name == 'nt':
            result = run(cmd_line, cwd=session_path,stdout=PIPE, check=False, shell=True)
        else:
            result = run([cmd_line], cwd=session_path, stdout=PIPE, check=False, shell=True)
    
        # Check the results of the process using the returncodes and console output
        if result.returncode != 0 and result.returncode != 95 and result.returncode != 3:
            # SIMsalabim raised an error, stop the program and return the error message on the UI.
            startMessage = False
            message = ''
            result_decode = result.stdout.decode('utf-8')

            if result.returncode >= 100:
                # A fatal (numerical) error occurred. Return errorcode and a standard error message.
                message = fatal_error_message(result.returncode)
            else:
                # Simsalabim raised an error. Read the console output for the details / error messaging. 
                # All linetypes after 'Program will be terminated, press enter.' are considered part of the error message
                for line_console in result_decode.split('\n'):
                    if startMessage is True:
                        # The actual error message. Since the error message can be multi-line, append each line.
                        message = message + line_console + '\n'
                    if 'Program will be terminated.' in line_console:
                        # Last 'regular' line of the console output. The next line is from the error message.
                        startMessage = True

            # Show the message as an error on the screen. Do not continue to the simulation results page.
            message = 'Simulation raised an error with Errorcode: ' + str(result.returncode) + '\n\n' + message
        else:
            # SIMsalabim simulation succeeded
            if result.returncode == 95:
                # In case of errorcode 95, failures during the simulations were encountered but the simulation did not halt. Show 'error' messages on the UI.
                startMessage = False
                message = ''
                result_decode = result.stdout.decode('utf-8')
                for line_console in result_decode.split('\n'):
                    if startMessage is True:
                        # The actual error message. Since the error message can be multi-line, append each line.
                        message = message + line_console + '\n'
                    if 'Program will be terminated.' in line_console:
                        # Last 'regular' line of the console output. The next line is from the error message.
                        startMessage = True
               
                # Show the message as a success on the screen
                message = 'Simulation completed but raised errorcode: ' + str(result.returncode) + '\n\n' + 'The simulation finished but at least 1 point did not converge. \n\n' + message
            elif result.returncode == 3:
                # Special case, should not occur in the web version.
                # When the program exits as a success but no simulation has been run, e.g. in the case of the autotidy functionality. 
                message = 'Action completed'
            else:
                # Simulation completed as expected.
                message = 'Simulation complete. Output can be found in the Simulation results.'
    else:
        # if verbose:
        if os.name == 'nt':
            result = run(cmd_line, cwd=session_path,stdout=PIPE, check=False, shell=True)
        else:
            result = run([cmd_line], cwd=session_path, stdout=PIPE, check=False, shell=True)
        
        if verbose:
            startMessage = False
            message = ''
            result_decode = result.stdout.decode('utf-8')
            for line_console in result_decode.split('\n'):
                if startMessage is True:
                    # The actual error message. Since the error message can be multi-line, append each line.
                    message = message + line_console + '\n'
                if 'Program will be terminated.' in line_console:
                    # Last 'regular' line of the console output. The next line is from the error message.
                    startMessage = True
            print(message)
        message = ''
    result = result.returncode
    return result, message


def run_simulation_filesafe(sim_type, cmd_pars, session_path, run_mode = False, verbose = False, **kwargs):
    """Run the SIMsalabim simulation executable with the chosen device parameters. 
        Return the complete result object of the process accompanied by a message with information, 
        in case of both success and failure.

    Parameters
    ----------
    sim_type : string
        Which type of simulation to run: simss or zimt
    cmd_pars : List
        List with parameters to add to the simss/zimt cmd line. Each parameter is a dict with par,val keys. 
        Note: when relevant the first entry must be the deviceparameters file with a key: dev_par_file
    session_path : string
        File path of the simss or zimt executable 
    run_mode : boolean
        True if function is called as part of The Shell, False when called directly. 
        Prevents using streamlit components outside of The Shell.
    verbose : boolean
        True if the console output of the simulation should be printed to the console
    **kwargs : dict
        Additional keyword arguments, not used in this function but can be used to pass additional parameters to the function.

    Returns
    -------
    result : int
        Return code of the simulation process, 0 for success, other values for errors.
    string
        Return message to display on the UI, for both success and failed
    """
    max_wait_time = kwargs.get('max_wait_time', 100)  # seconds
    # Create a temp folder to store the simulation results
    ID = str(uuid.uuid4())
    tmp_folder = os.path.join(session_path, 'tmp_'+ID)
    if not os.path.exists(tmp_folder):
        os.makedirs(tmp_folder)

    # get file setup from cmd_pars
    device_parameters = None

    for cmd_par in cmd_pars:
        if cmd_par['par'] == 'dev_par_file':
            device_parameters = cmd_par['val']
            break

    if device_parameters is None:
        raise ValueError('Device parameters file not found in the command parameters list.')

    # Copy the device parameters file to the temp folder
    make_thread_safe_file_copy(os.path.join(session_path, device_parameters), tmp_folder)

    # copy the executable to the temp folder
    if os.name == 'nt':
        shutil.copy(os.path.join(session_path, sim_type+'.exe'), tmp_folder)
    else:
        shutil.copy(os.path.join(session_path, sim_type), tmp_folder)

    device_parameters = os.path.basename(device_parameters)
    #
    while True:
        # copy the file to the temp folder
        try:
            dev_par, layers = load_device_parameters(session_path, device_parameters, run_mode = False)
            break
        except:
            pass 
        time.sleep(0.002)

    if os.name != 'nt':
        try:
            dev_par, layers = load_device_parameters(session_path, device_parameters, run_mode = False)
        except Exception as e:
            raise ValueError('Error loading device parameters check that all the input files are in the right directory. \n Error: {}'.format(e))
    else:
        warning_timeout = kwargs.get('warning_timeout', 10)
        exit_timeout = kwargs.get('exit_timeout', 60)
        t_wait = 0
        while True: # need this to be thread safe
            try:
                dev_par, layers = load_device_parameters(session_path, device_parameters, run_mode = False)
                break
            except Exception as e:
                time.sleep(0.002)
                t_wait = t_wait + 0.002
                if t_wait > warning_timeout:
                    print('Warning: SIMsalabim is not responding, please check that all the input files are in the right directory')
                if t_wait > exit_timeout:
                    raise ValueError('Error loading device parameters check that all the input files are in the right directory. \n Error: {}'.format(e))

    # check for new layers in the cmd_pars
    newlayers = []
    idx_layers = []
    for cmd_par in cmd_pars:
        if cmd_par['par'].startswith('l') and cmd_par['par'][1:].isdigit():
            newlayers.append(cmd_par)
            idx_layers.append(int(cmd_par['par'][1:]))
    
    # update layer files with cmd_pars
    if len(newlayers) > 0:
        for layer in layers:
            for cmd_par in cmd_pars:
                if cmd_par['par'] == layer[1]:
                    layer[2] = cmd_par['val']
                    break
    
    # check if we need to add new layers
    newlayers = [x for _, x in sorted(zip(idx_layers, newlayers))]
    idx_layers = sorted(idx_layers)
    if len(idx_layers) > 0:
        if max(idx_layers) > len(layers)-1: # means we need to add new layers
            # check that all the layers between len(layers)-& and max(idx_layers) are in the idx_layers list
            for i in range(len(layers),max(idx_layers)+1):
                if i not in idx_layers:
                    raise ValueError('Missing layer definition for layer '+str(i)+' in the command parameters list.')
            # add the new layers
            for i, layer in zip(idx_layers, newlayers):
                layers.append(['par','l'+str(i),layer['val']],'parameter file for layer '+str(i))
        
            # update dev_par layer files with cmd_pars
            dev_par_keys = list(dev_par.keys())
            for i, section in enumerate(dev_par[dev_par_keys[0]]):
                if section[0].lower() == 'layers':
                    # update section with layers[1:]
                    dev_par[dev_par_keys[0]][i][1:] = layers[1:]   


    # move all layer files to the temp folder
    for layer in layers:
        make_thread_safe_file_copy(os.path.join(session_path, layer[2]), tmp_folder)

    res = store_file_names(dev_par, sim_type, device_parameters, layers, run_mode = False)
    layer_files = res[0]
    optical_files = res[1]
    # make absolute paths
    layer_files = [os.path.abspath(os.path.join(session_path, f)) for f in layer_files]
    optical_files = [os.path.abspath(os.path.join(session_path, f)) for f in optical_files]
    optical_file_basenames = [os.path.basename(f) for f in optical_files]   
    traps_int_files = res[2]
    traps_bulk_files = res[3]
    traps_int_files = [os.path.abspath(os.path.join(session_path, f)) for f in traps_int_files]
    traps_bulk_files = [os.path.abspath(os.path.join(session_path, f)) for f in traps_bulk_files]
    traps_int_file_basenames = [os.path.basename(f) for f in traps_int_files]
    traps_bulk_file_basenames = [os.path.basename(f) for f in traps_bulk_files]

    ExpJV_file = None
    tVGFile = None
    tJFile = None
    JVFile = None
    scParsFile = None
    if sim_type == 'simss':
        if res[4].lower() == 'None'.lower():
            ExpJV_file = None
        else:
            ExpJV_file = res[4]
        JVFile = res[7]
        scParsFile = res[8]
    elif sim_type == 'zimt':
        tVGFile = res[4]
        tJFile = res[7]
    else :
        raise ValueError('Simulation type not recognized.')

    varFile = res[5]
    logFile = res[6] 
    
    # Copy the files to the temporary folder
    for file in layer_files + optical_files + traps_int_files + traps_bulk_files + [ExpJV_file] + [tVGFile]:
        if file is not None and os.path.isfile(file):
            make_thread_safe_file_copy(file, tmp_folder)
        
        # update temp folder files with basename 
        if file in layer_files:
            make_basename_input_files(os.path.join(tmp_folder, os.path.basename(file)))       
    
    input_files = get_inputFile_from_cmd_pars(sim_type, cmd_pars)
    
    input_files = [os.path.abspath(os.path.join(session_path, f['val'])) for f in input_files]
    input_files_basenames = [os.path.basename(f) for f in input_files]
    
    for file in input_files:
        # print(file)
        # if file already in the tmp_folder, remove it and copy the new one
        while True:
            try:
                if os.path.isfile(os.path.join(tmp_folder, os.path.basename(file))):
                    os.remove(os.path.join(tmp_folder, os.path.basename(file)))
                break
            except:
                pass
            # sleep random time to prevent high CPU usage smaller than 5ms
            time.sleep(random.uniform(0.002,0.005))
            
        # if os.path.isfile(os.path.join(tmp_folder, os.path.basename(file))):
        #     os.remove(os.path.join(tmp_folder, os.path.basename(file)))
        make_thread_safe_file_copy(os.path.join(session_path, file), tmp_folder)
    
    # set basename for device parameters
    make_basename_input_files(os.path.join(tmp_folder, os.path.basename(device_parameters)))

    # Construct the command to run the executable
    cmd_pars = make_basename_file_cmd_pars(cmd_pars)
    cmd_line = construct_cmd(sim_type, cmd_pars)

    # Run the simulation
    if os.name == 'nt':
        result = run(cmd_line, cwd=tmp_folder, stdout=PIPE, check=False, shell=True)
    else:
        result = run([cmd_line], cwd=tmp_folder, stdout=PIPE, check=False, shell=True)
    
    # Check the results of the process using the returncodes and console output
    if result.returncode != 0 and result.returncode != 95 and result.returncode != 3:
        # SIMsalabim raised an error, stop the program and return the error message on the UI.
        startMessage = False
        message = ''
        result_decode = result.stdout.decode('utf-8')

        if result.returncode >= 100:
            # A fatal (numerical) error occurred. Return errorcode and a standard error message.
            message = fatal_error_message(result.returncode)
        else:
            # Simsalabim raised an error. Read the console output for the details / error messaging. 
            # All linetypes after 'Program will be terminated, press enter.' are considered part of the error message
            for line_console in result_decode.split('\n'):
                if startMessage is True:
                    # The actual error message. Since the error message can be multi-line, append each line.
                    message = message + line_console + '\n'
                if 'Program will be terminated.' in line_console:
                    # Last 'regular' line of the console output. The next line is from the error message.
                    startMessage = True

        # Show the message as an error on the screen. Do not continue to the simulation results page.
        message = 'Simulation raised an error with Errorcode: ' + str(result.returncode) + '\n\n' + message
    else:
        # SIMsalabim simulation succeeded
        if result.returncode == 95:
            # In case of errorcode 95, failures during the simulations were encountered but the simulation did not halt. Show 'error' messages on the UI.
            startMessage = False
            message = ''
            result_decode = result.stdout.decode('utf-8')
            for line_console in result_decode.split('\n'):
                if startMessage is True:
                    # The actual error message. Since the error message can be multi-line, append each line.
                    message = message + line_console + '\n'
                if 'Program will be terminated.' in line_console:
                    # Last 'regular' line of the console output. The next line is from the error message.
                    startMessage = True

            # Show the message as a success on the screen
            message = 'Simulation completed but raised errorcode: ' + str(result.returncode) + '\n\n' + 'The simulation finished but at least 1 point did not converge. \n\n' + message
        elif result.returncode == 3:
            # Special case, should not occur in the web version.
            # When the program exits as a success but no simulation has been run, e.g. in the case of the autotidy functionality. 
            message = 'Action completed'
        else:
            # Simulation completed as expected.
            message = 'Simulation complete. Output can be found in the Simulation results.'

    # Move output files to the original folder
    if sim_type.lower() == 'simss':
        if os.path.isfile(os.path.join(tmp_folder, varFile)):
            # shutil.move(os.path.join(tmp_folder, varFile), os.path.join(session_path, varFile))
            make_thread_safe_file_copy(os.path.join(tmp_folder, varFile), session_path)
        if os.path.isfile(os.path.join(tmp_folder, logFile)):
            # shutil.move(os.path.join(tmp_folder, logFile), os.path.join(session_path, logFile))
            make_thread_safe_file_copy(os.path.join(tmp_folder, logFile), session_path)
        if os.path.isfile(os.path.join(tmp_folder, JVFile)):
            # shutil.move(os.path.join(tmp_folder, JVFile), session_path)
            make_thread_safe_file_copy(os.path.join(tmp_folder, JVFile), session_path)
        if os.path.isfile(os.path.join(tmp_folder, scParsFile)):
            # shutil.move(os.path.join(tmp_folder, scParsFile), os.path.join(session_path, scParsFile))
            make_thread_safe_file_copy(os.path.join(tmp_folder, scParsFile), session_path)
    elif sim_type.lower() == 'zimt':
        if os.path.isfile(os.path.join(tmp_folder, varFile)):
            # shutil.move(os.path.join(tmp_folder, varFile), os.path.join(session_path, varFile))
            make_thread_safe_file_copy(os.path.join(tmp_folder, varFile), session_path)
        if os.path.isfile(os.path.join(tmp_folder, logFile)):
            # shutil.move(os.path.join(tmp_folder, logFile), os.path.join(session_path, logFile))
            make_thread_safe_file_copy(os.path.join(tmp_folder, logFile), session_path)
        if os.path.isfile(os.path.join(tmp_folder, tVGFile)):
            # shutil.move(os.path.join(tmp_folder, tVGFile), os.path.join(session_path, tVGFile))
            make_thread_safe_file_copy(os.path.join(tmp_folder, tVGFile), session_path)
        if os.path.isfile(os.path.join(tmp_folder, tJFile)):
            # shutil.move(os.path.join(tmp_folder, tJFile), os.path.join(session_path, tJFile))
            make_thread_safe_file_copy(os.path.join(tmp_folder, tJFile), session_path)

    result = result.returncode
    return result, message


    
def make_thread_safe_file_copy(file, destination):
    """Copy a file to a temp folder, and wait until the file is not in use anymore.

    Parameters
    ----------
    file : string
        File path of the file to copy
    destination : string
        File path of the destination folder
    """
    
    # check temp folder exists if not create it
    if not os.path.exists(destination):
        os.makedirs(destination)

    # tries to open the file, if it is in use it will wait until it is not
    while True:
        
        # copy the file to the temp folder
        try:
            shutil.copy(file, destination)
            break
        except:
            pass
        
        # add 2ms delay to prevent high CPU usage
        time.sleep(0.002)



""" Cleaning output files """
######### Package Imports ###################################################################

import os, shutil

######### Function Definitions ##############################################################

def clean_up_output(filename_start,path):
    """Delete output files from the simulation

    Parameters
    ----------
    filename_start : string
        string containing the beginning of the filename to delete
    path : string
        path to the directory where we clean the output
    """ 
    for fname in os.listdir(path):
        if fname.startswith(filename_start) and not os.path.isdir(os.path.join(path,fname)):
            os.remove(os.path.join(path,fname))

def store_output_in_folder(filenames,folder_name,path):
    """Move output files from the simulation into new folder

    Parameters
    ----------
    filenames : list of string
        list of string containing the name of the files to move
    folder_name : string
        name of the folder where we store the output files       
    path : string
        directory of the folder_name (creates one if it does not already exist)
    """    

    # Create directory if it does not exist
    if not os.path.exists(os.path.join(path,folder_name)):
        os.makedirs(os.path.join(path,folder_name))
    # move file into the new folder
    for i in filenames:
        if os.path.exists(os.path.join(path,i)):
            os.replace(os.path.join(path,i),os.path.join(path,folder_name,i))
        else:
            print('File {} does not exist'.format(os.path.join(path,i)))

def clean_file_type(ext,path):
    """Delete files of a given type in the current directory

    Parameters
    ----------
    ext : string
        extension of the files to delete 
    path : string
        path to the directory where we clean the output

    """ 
    for fname in os.listdir(path):
        if fname.endswith(ext):
            os.remove(os.path.join(path,fname))

def clean_all_output(path,filename_starts = ['JV','Var','tj','tVG','scPars','Str4Parallel','log'],exts = ['.png']):
    """Delete all files in the current directory

    Parameters
    ----------
    path : string
        path to the directory where we clean the output
    filename_starts : list of string
        list of string containing the beginning of the filename to delete, default is ['JV','Var','tj','tVG','scPars','Str4Parallel','log']
    exts : list of string
        list of string containing the extension of the files to delete, default is ['.png']
    
    """ 
    

    for filename_start in filename_starts:
        clean_up_output(filename_start,path)

    for ext in exts:
        clean_file_type(ext,path)

def delete_folders(folder_name_start,path):
    """Delete all folders starting with a given name in the current directory, even if they are not empty

    Parameters
    ----------
    folder_name : string
        name of the folder to delete
    path : string
        path to the directory where we delete the folder

    """ 
    for fname in os.listdir(path):
        folder_path = os.path.join(path, fname)
        if fname.startswith(folder_name_start) and os.path.isdir(folder_path):
            shutil.rmtree(folder_path)
            



"""Check and convert Windows paths to long path format if necessary."""
# Description:
# ------------
# This script checks if a given Windows path exceeds the traditional MAX_PATH limit
# and converts it to a long path format if necessary.

######### Package Imports ###################################################################

from pathlib import Path

######### Function Definitions ##############################################################

def convert_to_long_path(tempPath):
    """ Convert a path to a Windows long path by adding the \\\\?\\ prefix

    Parameters
    ----------
    tempPath : string
        Path to be converted

    Returns
    -------
    Path
        Converted Path object with long path prefix
    """
    resolved_path = Path(tempPath).resolve()
    if len(str(resolved_path)) < 260: # Windows path length limit
        return resolved_path 
    
    else:
        print(f"\n Warning: Path is too long ({len(resolved_path)} characters), if the process fails try to shorten the path (i.e. by moving the folder to a location with a shorter path). \n If it stills fails try changing the Windows registry to enable long paths: https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation?tabs=registry#enabling-long-paths-in-windows-10-version-1607-and-later \n Current Path: {tempPath}")

        # Add \\?\ prefix manually for long paths
        win_path = Path("\\\\?\\" + str(resolved_path))
        return win_path
    


""" Test EQE module with pySIMsalabim"""

######### Package Imports #########################################################################
import os, sys, uuid 
import pandas as pd
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.experiments.EQE import *

######### Test Functions #########################################################################

def test_run_EQE():
    """ Test the run_EQE function """
    # Set the path to the simulation setup file
    if os.path.exists('SIMsalabim'):
        cwd = os.path.abspath('.')
    else:
        cwd = os.path.abspath('../..')
    simss_device_parameters = os.path.join(cwd, 'SIMsalabim','SimSS','simulation_setup.txt')
    session_path = os.path.join(cwd, 'SIMsalabim','SimSS')
    spectrum = os.path.join(cwd, 'SIMsalabim','Data','AM15G.txt')
    lambda_min = 300
    lambda_max = 800
    lambda_step = 10
    Vext = 0
    outfile_name = 'EQE.dat'
    sim_type = 'simss'
    # Run the EQE simulation
    ret, mess = run_EQE(simss_device_parameters, session_path, spectrum, lambda_min, lambda_max, lambda_step, Vext, outfile_name = outfile_name, JV_file_name = 'JV.dat', run_mode = False, parallel = False, force_multithreading = False, UUID=str(uuid.uuid4()), cmd_pars=[{'par': 'l2.L', 'val': '50e-9'}],threadsafe=False)
    
    # Clean up the output
    sim.clean_all_output(session_path)
    sim.clean_up_output('EQE',session_path)
    sim.delete_folders('tmp',session_path)
    # Check the output
    assert ret == 0, 'EQE simulation failed'

def test_EQE_parallel():
    """ Test the EQE_parallel function """
    # try:
    if os.path.exists('SIMsalabim'):
        cwd = os.path.abspath('.')
    else:
        cwd = os.path.abspath('../..')
    session_path = os.path.join(cwd, 'SIMsalabim','SimSS')

    def run(cmd_pars,ID):
        simss_device_parameters = os.path.join(cwd, 'SIMsalabim','SimSS','simulation_setup.txt')
        session_path = os.path.join(cwd, 'SIMsalabim','SimSS')
        spectrum = os.path.join(cwd, 'SIMsalabim','Data','AM15G.txt')
        lambda_min = 300
        lambda_max = 800
        lambda_step = 10
        Vext = 0
        outfile_name = 'EQE.dat'
        print('Running')
        ret, mess = run_EQE(simss_device_parameters, session_path, spectrum, lambda_min, lambda_max, lambda_step, Vext, outfile_name = outfile_name, JV_file_name = 'JV.dat', run_mode = False, parallel = True, force_multithreading = True, UUID=ID, cmd_pars=cmd_pars,max_jobs=4, threadsafe = True)

    ID1 = str(uuid.uuid4())
    ID2 = str(uuid.uuid4())
    cmd_pars = [{'par': 'l2.L', 'val': '50e-9'}]
    cmd_pars2 = [{'par': 'l2.L', 'val': '300e-9'}]

    # wrap it in joblib to run in parallel
    ID_list = [ID1,ID2]
    cmd_list = [cmd_pars,cmd_pars2]
    Parallel(n_jobs=2)(delayed(run)(cmd_list[i],ID_list[i]) for i in range(2))

    # print('Plotting')
    # plt.figure()
    # df = pd.read_csv(os.path.join(session_path,'EQE_'+ID1+'.dat'), sep = r'\s+')
    # plt.plot(df['lambda'],df['EQE'])

    # # plt.figure
    # df2 = pd.read_csv(os.path.join(session_path,'EQE_'+ID2+'.dat'), sep = r'\s+')
    # plt.plot(df2['lambda'],df2['EQE'])
    # plt.show()

    sim.clean_all_output(session_path)
    sim.clean_up_output('EQE',session_path)
    sim.delete_folders('tmp',session_path)
    assert True
    # except Exception as e:
    #     print(e)
    #     assert False, 'EQE parallel simulation failed'

if __name__ == '__main__':
    test_run_EQE()
    test_EQE_parallel()
    print('All EQE tests passed')




import os
import importlib.util
import pkgutil

# Path to the directory of this file
package_dir = os.path.dirname(__file__)

# Loop through all modules in the current directory
for (_, module_name, _) in pkgutil.iter_modules([package_dir]):
    # Construct the full path to the module
    module_path = os.path.join(package_dir, module_name + '.py')
    # Import the module
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    # Add all functions from the module to globals() so they can be imported directly
    for attribute_name in dir(module):
        attribute = getattr(module, attribute_name)
        if callable(attribute):
            globals()[attribute_name] = attribute


# from pySIMsalabim.utils import *

import os, sys, warnings

from . import utils
from .utils import clean_up, device_parameters, general, parallel_sim, utils
from .utils.clean_up import *
from .utils.device_parameters import *
from .utils.general import *
from .utils.parallel_sim import *
from .utils.utils import *

from . import plots
from .plots import plot_def, plot_functions, band_diagram
from .plots.plot_def import *
from .plots.plot_functions import *
from .plots.band_diagram import *

from . import aux_funcs
from .aux_funcs import addons
from .aux_funcs.addons import *



"""Perform Intensity-Modulated Photocurrent Spectroscopy (IMPS) simulations"""
######### Package Imports #########################################################################

import os, sys
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import math
import scipy.integrate
# import pySIMsalabim
## Import pySIMsalabim, if not successful, add the parent directory to the system path
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.utils import general as utils_gen
from pySIMsalabim.plots import plot_functions as utils_plot
from pySIMsalabim.utils.utils import *
from pySIMsalabim.utils.device_parameters import *

######### Function Definitions ####################################################################

def create_tVG_IMPS(V, G_frac, del_G, tVG_name, session_path, f_min, f_max, ini_timeFactor, timeFactor):
    """Create a tVG file for IMPS experiments.

    Parameters
    ----------
    V : float
        Voltage, the voltage is constant over the whole time range
    G_frac : float
        Fractional light intensity
    del_G : float
        Applied generation rate increase at t=0
    tVG_name : string
        Name of the tVG file
    session_path : string
        Path of the simulation folder for this session
    f_min : float
        Minimum frequency
    f_max : float
        Maximum frequency
    ini_timeFactor : float
        Constant defining the size of the initial timestep
    timeFactor : float
        Exponential increase of the timestep, to reduce the amount of timepoints necessary. Use values close to 1.

    Returns
    -------
    string
        A message to indicate the result of the process
    """
    
    time = 0
    del_t = ini_timeFactor/f_max

    # Starting line of the tVG file: header + datapoints at time=0. Set the correct header
    tVG_lines = 't Vext G_frac\n' + f'{time:.3e} {V} {G_frac:.3e}\n'

    # Make the other lines in the tVG file
    while time < 1/f_min: #max time: 1/f_min is enough!
        time = time+del_t
        tVG_lines += f'{time:.3e} {V} {G_frac+del_G:.3e}\n'
        del_t = del_t * timeFactor # At first, we keep delt constant to its minimum values

    # Export tVG file to ZimT folder
    with open(os.path.join(session_path,tVG_name), 'w') as file:
        file.write(tVG_lines)

    # tVG file is created, message a success
    msg = 'Success'
    retval = 0
    return retval, msg

# def read_tj_file(session_path, tj_file_name='tj.dat'):
#     """ Read relevant parameters for admittance of the tj file

#     Parameters
#     ----------
#     session_path : string
#         Path of the simulation folder for this session
#     data_tj : dataFrame
#         Pandas dataFrame containing the tj output file from ZimT

#     Returns
#     -------
#     DataFrame
#         Pandas dataFrame of the tj_file containing the time, current density, numerical error in the current density and the photogenerated current density
#     """

#     data = pd.read_csv(os.path.join(session_path,tj_file_name), sep=r'\s+')

#     return data

# def get_integral_bounds(data, f_min=1e-2, f_max=1e6, f_steps=20):
#     """ Determine integral bounds in the time domain, used to compute the conductance and capacitance

#     Parameters
#     ----------
#     data : dataFrame
#         Pandas dataFrame containing the time, current density, numerical error in the current density and the photogenerated current density of the tj_file
#     f_min : float
#         Minimum frequency
#     f_max : float
#         Maximum frequency
#     f_steps : float
#         Frequency steps

#     Returns
#     -------
#     list
#         List of array indices that will be used in the plotting
#     """

#     # Total number of time points
#     numTimePoints = len(data['t'])

#     # Check which time index corresponds to 1/fmax. We call this istart:
#     istart = -1
#     for i in range(numTimePoints):
#         if math.isclose(data['t'][i], 1/f_max, rel_tol = 2/f_steps): #note: don't use == to compare 2 floating points!
#             istart = i

#     # Starting time point could not be found
#     if istart == -1:
#         msg = 'Could not find a time that corresponds to the highest frequency.'
#         return -1, msg
    
#     # print('Found istart: ', istart)

#     # ifin: last index we should plot, corresponds to time = 1/f_min:
#     ifin = numTimePoints - 1

#     # isToPlot starts with istart:
#     isToPlot = [istart]

#     PlotRatio = max(1, round( (ifin-istart)/(math.log10(f_max/f_min) * f_steps)))

#     # Incorrect plot ratio
#     if PlotRatio < 1:
#         msg = 'PlotRatio smaller than 1. It should at least be 1'
#         return -1, msg

#     # Then add the other indices:
#     for i in range(istart+1, ifin-1):
#         if (i-istart) % PlotRatio == 0: # note: % is python's modulo operator.
#             isToPlot.append(i) # add the index to our array

#     # Also include the last index:
#     isToPlot.append(ifin)

#     # Integral bounds have been determined, return the array with indices and a success message
#     msg = 'Success'
#     return isToPlot, msg

def calc_IMPS_limit_time(I, errI, time, imax):
    """Fourier Decomposition formula which computes the admittance at frequency freq (Hz) and its complex error
    Based on S.E. Laux, IEEE Trans. Electron Dev. 32 (10), 2028 (1985), eq. 5a, 5b
    We integrate from 0 to time[imax] at frequency 1/time[imax]

    Parameters
    ----------
    I : np.array
        Array of currents
    errI : np.array
        Numerical error in calculated currents (output of ZimT)
    time : np.array
        Array with all time positions, from 0 to tmax
    imax : integer
        Index of the last timestep/first frequency for which the integrals are calculated

    Returns
    -------
    float
        Frequency belonging to Y(f)
    complex number
        Admittance at frequency f: Y(f)
    complex number
        Numerical error in calculated admittance Y(f)
    """

    freq=1/time[imax] #we obtain the frequency from the time array
    Iinf = I[imax] # I at infinite time, i.e. the last one we have.
	
    #prepare array for integrants:
    int1 = np.empty(imax)
    int2 = np.empty(imax)
    int3 = np.empty(imax)
    int4 = np.empty(imax)
	
    #now we use only part of the time array:
    timeLim = time[0:imax]
	
    for i in range(imax) :
        sinfac = math.sin(2*math.pi*freq*timeLim[i])
        cosfac = math.cos(2*math.pi*freq*timeLim[i])
        int1[i] = sinfac*(I[i] - Iinf)
        int2[i] = cosfac*(I[i] - Iinf)	
        int3[i] = sinfac*(I[i] + errI[i] - Iinf - errI[imax])
        int4[i] = cosfac*(I[i] + errI[i] - Iinf - errI[imax])	

    #now compute the conductance and capacitance:
    cond = (Iinf - I[0] + 2*math.pi*freq*scipy.integrate.trapezoid(int1, timeLim))
    cap = scipy.integrate.trapezoid(int2, timeLim)
    #convert to admittance:
    Y = cond + 2J*math.pi*freq*cap
	
    #and again, but now with the error added to the current:	
    condErr = (Iinf + errI[imax] - I[0] - errI[0] + 2*math.pi*freq*scipy.integrate.trapezoid(int3, timeLim))
    capErr = scipy.integrate.trapezoid(int4, timeLim)
    #convert to admittance:
    Y2 = condErr + 2J*math.pi*freq*capErr
    
    #error is the difference between Y and Y2:
    errY = Y - Y2
    
    #now return complex admittance, its error and the corresponding frequency:	
    return freq, Y, errY

def calc_IMPS(data, isToPlot):
    """ Calculate the admittance over the frequency range
    
    Parameters
    ----------
    data : dataFrame
        Pandas dataFrame containing the time, current density, numerical error in the current density and the photogenerated current density of the tj_file
    isToPlot : list
        List of array indices that will be used in the plotting

    Returns
    -------
    np.array
        Array of frequencies
    np.array
        Array of the real component of admittance
    np.array
        Array of the imaginary component of admittance
    np.array
        Array of complex error
    """
    # init the arrays for the admittance and its error:
    numFreqPoints = len(isToPlot)
    freq = np.empty(numFreqPoints)
    ReY = np.empty(numFreqPoints)
    ImY = np.empty(numFreqPoints)
    Y = [1 + 1J] * numFreqPoints
    errY = [1 + 1J] * numFreqPoints

    for i in range(numFreqPoints):
        imax=isToPlot[i]
        freq[i], Y[i], errY[i] = calc_IMPS_limit_time(data['Jext'], data['errJ'], data['t'], imax)
        # we are only interested in the absolute value of the real and imag components:
        ReY[i] =(Y[i].real)
        ImY[i] = (Y[i].imag)
    
    return freq, ReY, ImY, errY

def store_IMPS_data(session_path, freq, ReY, ImY, errY, output_file):
    """ Save the frequency, real & imaginary part of the admittance & its error in one file called freqY.dat
    
    Parameters
    ----------
    session_path : string
        working directory for zimt
    freq : np.array
        Array of frequencies
    ReY : np.array
        Array of the real component of admittance
    ImY : np.array
        Array of the imaginary component of admittance
    errY : np.array
        Array of complex error in admittance
    """

    with open(os.path.join(session_path,output_file), 'w') as file:
        file.write('freq ReY ImY ReErrY ImErrY' + '\n')
        for i in range(len(freq)):
            file.write(f'{freq[i]:.6e} {ReY[i]:.6e} {ImY[i]:.6e} {abs(errY[i].real):.6e} {abs(errY[i].imag):.6e}' + '\n')

    # print('The data of the IMPS graphs is written to ' + output_file)

def get_IMPS(data, f_min, f_max, f_steps, session_path, output_file):
    """Calculate the IMPS from the simulation result

    Parameters
    ----------
    data : DataFrame
        DataFrame with the simulation results (tj.dat) file
    f_min : float
        Minimum frequency
    f_max : float
        Maximum frequency
    f_steps : float
        Frequency step
    session_path : string
        working directory for zimt
    output_file : string
        name of the file where the IMPS data is stored

    Returns
    -------
    integer,string
        returns -1 (failed) or 1 (success), including a message
    """
    isToPlot, msg = get_integral_bounds(data, f_min, f_max, f_steps)

    if isToPlot != -1:
        # Integral bounds have been determined, continue to calculate the IMPS
        freq, ReY, ImY, errY = calc_IMPS(data, isToPlot)

        # Write IMPS results to a file
        store_IMPS_data(session_path, freq, ReY, ImY, errY, output_file)

        msg = 'Success'
        return 0, msg
    else:
        # Failed to determine integral bounds, exit with the error message
        return -1, msg

def IMPS_plot(session_path, output_file, xscale='log', yscale='log', plot_type = plt.plot):
    """ Plot the real and imaginary part of the admittance against frequency

    Parameters
    ----------
    session_path : string
        working directory for zimt
    xscale : string
        Scale of the x-axis. E.g linear or log
    yscale_ax1 : string
        Scale of the left y-axis. E.g linear or log
    yscale_ax2 : string
        Scale of the right y-axis. E.g linear or log
    """
    # Read the data from freqY-file
    data_freqY = pd.read_csv(os.path.join(session_path,output_file), sep=r'\s+')

    # Flip the ImY data to the first quadrant
    # data["ImY"] = data["ImY"]*-1*-1

    # Define the plot parameters, two y axis
    pars_imps = {'ImY' : '-Im Y [A/m$^2$]' }
    selected = ['ImY']
    par_x = 'freq'
    xlabel = 'frequency [Hz]'
    ylabel = 'Im Y'
    title = 'Computed IMPS'
    fig, ax = plt.subplots()
    utils_plot.plot_result(data_freqY, pars_imps, selected, par_x, xlabel, ylabel, xscale, yscale, title, ax, plot_type)

    plt.show()

def ColeCole_plot(session_path, output_file, xscale='linear', yscale='linear', plot_type = plt.errorbar):
    """ Plot the Cole-Cole plot with the real and imaginary part of the admittance against frequency

    Parameters
    ----------
    session_path : string
        working directory for zimt
    output_file : string
        Filename where the admittance data is stored
    xscale : string
        Scale of the x-axis. E.g linear or log
    yscale : string
        Scale of the y-axis. E.g linear or log
    plot_type : matplotlib.pyplot
        Type of plot to display
    """
    # Read the data from freqY-file
    data = pd.read_csv(os.path.join(session_path,output_file), sep=r'\s+')
    
    fig, ax = plt.subplots()
    pars_nyq = {'ImY' : '-Im Y [A/m$^2$]'}
    par_x_nyq = 'ReY'
    par_weight_nyq = 'freq'
    xlabel_nyq = 'Re Y [A/m$^2$]'
    ylabel_nyq = '-Im Y [A/m$^2$]'
    weightlabel_nyq = 'frequency [Hz]'
    weight_norm_nyq = 'log'
    title_nyq = 'Cole-Cole plot'

    # Plot the Cole-Cole plot with or without errorbars
    if plot_type == plt.errorbar:
        ax = utils_plot.plot_result(data, pars_nyq, list(pars_nyq.keys()), par_x_nyq, xlabel_nyq, ylabel_nyq, xscale, yscale, title_nyq, ax, plot_type, 
                                            [], data['ImErrY'], legend=False)
    else:
        ax = utils_plot.plot_result(data, pars_nyq, list(pars_nyq.keys()), par_x_nyq, xlabel_nyq, ylabel_nyq, xscale, yscale, title_nyq, ax, plot_type, legend=False)

    plt.show()



def plot_IMPS(session_path, output_file='freqY.dat'):
    """Plot admittance of IMPS

    Parameters
    ----------
    session_path : string
        working directory for zimt
    """
    # IMPS plot
    IMPS_plot(session_path,output_file)

    # Cole-Cole plot
    ColeCole_plot(session_path,output_file)

def run_IMPS_simu(zimt_device_parameters, session_path, f_min, f_max, f_steps, V, G_frac, GStep = 0.05, run_mode=False, tVG_name = 'tVG.txt', output_file = 'freqY.dat', tj_name = 'tj.dat',varFile ='none', ini_timeFactor=1e-3, timeFactor=1.02, **kwargs):
    """Create a tVG file and run ZimT with admittance device parameters

    Parameters
    ----------
    zimt_device_parameters : string
        Name of the zimt device parameters file
    session_path : string
        Path of the simulation folder for this session
    f_min : float
        Minimum frequency
    f_max : float
        Maximum frequency
    f_steps : float
        Frequency step
    V : float
        Voltage, the voltage is constant over the whole time range
    G_frac : float
        Fractional light intensity
    GStep : float, optional
        Applied generation rate increase at t=0, by default 0.05
    run_mode : bool, optional
        Indicate whether the script is in 'web' mode (True) or standalone mode (False). Used to control the console output, by default False
    tVG_name : string, optional
        Name of the tVG file, by default tVG.txt
    output_file : string, optional
        Name of the file where the admittance data is stored, by default freqY.dat
    tj_name : string, optional
        Name of the tj file where the admittance data is stored, by default tj.dat
    varFile : string, optional
        Name of the var file, by default 'none'
    ini_timeFactor : float, optional
        Constant defining the size of the initial timestep, by default 1e-3
    timeFactor : float, optional
        Exponential increase of the timestep, to reduce the amount of timepoints necessary. Use values close to 1., by default 1.02
       
    Returns
    -------
    CompletedProcess
        Output object of with returncode and console output of the simulation
    string
        Return message to display on the UI, for both success and failed
    """
    verbose = kwargs.get('verbose', False) # Check if the user wants to see the console output
    UUID = kwargs.get('UUID', '') # Check if the user wants to add a UUID to the tj file name
    cmd_pars = kwargs.get('cmd_pars', None) # Check if the user wants to add additional command line parameters
    # Check if the user wants to force the use of thread safe mode, necessary for Windows with parallel simulations
    if os.name == 'nt':
        threadsafe = kwargs.get('threadsafe', True) # Check if the user wants to force the use of threads instead of processes
    else:
        threadsafe = kwargs.get('threadsafe', False) # Check if the user wants to force the use of threads instead of processes
    
    turnoff_autoTidy = kwargs.get('turnoff_autoTidy', None) # Check if the user wants to turn off the autoTidy function in SIMsalabim
    if turnoff_autoTidy is None: 
        if not threadsafe:
            turnoff_autoTidy = True
        else:
            turnoff_autoTidy = False

    # Update the file names with the UUID
    if UUID != '':
        dum_str = f'_{UUID}'
    else:
        dum_str = ''

    # Update the filenames with the UUID
    tj_name = os.path.join(session_path, tj_name)
    output_file = os.path.join(session_path, output_file)
    tVG_name = os.path.join(session_path, tVG_name)
    if UUID != '':
        tj_file_name_base, tj_file_name_ext = os.path.splitext(tj_name)
        tj_name = tj_file_name_base + dum_str + tj_file_name_ext 
        tVG_name_base, tVG_name_ext = os.path.splitext(tVG_name)
        tVG_name = tVG_name_base + dum_str + tVG_name_ext
        output_file_base, output_file_ext = os.path.splitext(output_file)
        output_file = output_file_base + dum_str + output_file_ext
        if varFile != 'none':
            var_file_base, var_file_ext = os.path.splitext(varFile)
            varFile = var_file_base + dum_str + var_file_ext
            varFile = os.path.join(session_path,varFile)
    # varFile = 'none' # we don't use a var file for this simulation

    # Create tVG
    result, message = create_tVG_IMPS(V, G_frac, GStep, tVG_name, session_path, f_min, f_max, ini_timeFactor, timeFactor)

    # Check if tVG file is created
    if result == 0:
        # Define mandatory options for ZimT to run well with IMPS:
        IMPS_args = [{'par':'dev_par_file','val':zimt_device_parameters},
                        {'par':'tVGFile','val':tVG_name},
                        {'par':'limitDigits','val':'0'},
                        {'par':'currDiffInt','val':'2'},
                        {'par':'tJFile','val':tj_name},
                        {'par':'varFile','val':varFile},
                        {'par':'logFile','val':'log'+dum_str+'.txt'}]
        
        if turnoff_autoTidy:
            IMPS_args.append({'par':'autoTidy','val':'0'})
        
        if cmd_pars is not None:
            IMPS_args = update_cmd_pars(IMPS_args, cmd_pars)

        if threadsafe:
            result, message = utils_gen.run_simulation_filesafe('zimt', IMPS_args, session_path, run_mode, verbose=verbose)
        else:
            result, message = utils_gen.run_simulation('zimt', IMPS_args, session_path, run_mode, verbose=verbose)

        if result == 0 or result == 95:
            data = read_tj_file(session_path, tj_file_name=tj_name)

            result, message = get_IMPS(data, f_min, f_max, f_steps, session_path, output_file)
            return result, message

        else:
            return result, message

    return result, message

## Running the function as a standalone script
if __name__ == "__main__":
    # IMPS input parameters
    f_min = 1e-2 # org 1e-2
    f_max = 5e6# org 1e6
    f_steps = 30 # org 30
    V_0 = 1.0 # Float or 'oc' for the open-circuit voltage
    G_frac = 1
    fac_G = 5e-2 # org 2e-1, use around 0.2 for IMPS

    # Define folder and file paths
    session_path = os.path.join('../../','SIMsalabim','ZimT')

    zimt_device_parameters = 'simulation_setup.txt'

    tVG_name = 'tVG.txt'
    tj_name = 'tj.dat'
    output_name = 'freqZ.dat'

    # UUID = str(uuid.uuid4()) # Add a UUID to the simulation
    UUID = ''

    # Not user input
    ini_timeFactor = 1e-3 # Initial timestep factor, org 1e-3
    timeFactor = 1.02 # Increase in timestep every step to reduce the amount of datapoints necessary, use value close to 1 as this is best! Org 1.02
    run_mode = False  # If False, show verbose output in console

    ############## Command line arguments  ##############
    ## Notes
    ## - The command line arguments are optional and can be provided in any order
    ## - Each command line argument must be provided in the format -par_name value
    ## - Possible arguments include all SIMsalabim parameters and IMPS specific parameters as listed before
    ## - Special arguments
    ##   - sp : string
    ##     - The session path, i.e. the working directory for the simulation
    ##   - simsetup : string
    ##     - The name of the zimt simulation setup parameters file
    ##   - UUID : string
    ##     - An UUID to add to the simulation (output)

    cmd_pars_dict = {}
    cmd_pars = []

    # Check if any arguments are provided.
    if len(sys.argv) >= 2:
        # Each arguments should be in the format -par val, i.e. in pairs of two. Skip the first argument as this is the script name
        if not len(sys.argv[1:]) % 2 == 0:
            print('Error in command line parameters. Please provide arguments in the format -par_name value -par_name value ...')
            sys.exit(1)
        else:
            input_list = sys.argv[1:]
            # Loop over the input list and put pairs into a dictionary with the first argument as key and the second argument as value
            for i in range(0,len(input_list),2):
                # Check if the key already exists, if not, add it to the cmd_pars_dict
                if str(input_list[i][1:]) in cmd_pars_dict:
                    print(f'Duplicate parameter found in the command line parameters: {str(input_list[i][1:])}')
                    sys.exit(1)
                else:
                    cmd_pars_dict[str(input_list[i][1:])] = str(input_list[i+1])

    # Check and process specific keys/arguments that are not native SIMsalabim arguments
    # Handle the session_path/sp argument separately, as the other parameters depend on this
    if 'sp' in cmd_pars_dict:
        session_path = cmd_pars_dict['sp']
        # remove from cmd_pars
        cmd_pars_dict.pop('sp')

    # Define mappings for keys and variables
    key_action_map = {
        'simsetup': lambda val: {'zimt_device_parameters': val},
        'f_min': lambda val: {'f_min': float(val)},
        'f_max': lambda val: {'f_max': float(val)},
        'f_steps': lambda val: {'f_steps': int(val)},
        'V_0': lambda val: {'V_0': float(val)},
        'G_frac': lambda val: {'G_frac': float(val)},
        'fac_G': lambda val: {'fac_G': float(val)},
        'tVG_name': lambda val: {'tVG_name': val},
        'tj_name': lambda val: {'tj_name': val},
        'out_name': lambda val: {'output_name': val},
        'UUID': lambda val: {'UUID': val},
    }

    for key in list(cmd_pars_dict.keys()):  # Use list to avoid modifying the dictionary while iterating
        if key in key_action_map:
            # Apply the corresponding action
            result = key_action_map[key](cmd_pars_dict[key])
            globals().update(result)  # Dynamically update global variables
            cmd_pars_dict.pop(key)

    # Handle remaining keys in `cmd_pars_dict` and add them to the cmd_pars list
    cmd_pars.extend({'par': key, 'val': value} for key, value in cmd_pars_dict.items())


    # Run IMPS spectroscopy
    GStep = G_frac*fac_G

    # result, message = run_IMPS_simu(zimt_device_parameters, f_min, f_max, f_steps, V, G_frac, GStep, tVG_name=tVG_name,  session_path= session_path, run_mode=False, ini_timeFactor=ini_timeFactor, timeFactor=timeFactor)
    result, message = run_IMPS_simu(zimt_device_parameters,session_path, f_min, f_max, f_steps, V_0, G_frac, GStep, run_mode=False, tVG_name=tVG_name, 
                                        output_file = output_name, tj_name = tj_name, ini_timeFactor=ini_timeFactor, timeFactor=timeFactor, cmd_pars=cmd_pars, UUID=UUID)

    # Make the IMPS plots
    if result == 0 or result == 95:
        plot_IMPS(session_path, os.path.basename(output_name))
    else:
        print(message)
        sys.exit(1)

